"""
This module contains all of the plotting functions for the JWST Rogue Path Tool.
Plotting output include:

* Valid angles at the exposure level
* Valid angles at the observation level
* Susceptibility region footprint w source intensities.
* V3 position angles vs flux at the observation level.

Authors
-------
    - Mario Gennaro
    - Mees Fix
"""

import astropy.units as u
from astropy.coordinates import SkyCoord
from matplotlib.patches import PathPatch, Wedge
from matplotlib.path import Path
import matplotlib.pyplot as plt
import numpy as np
from pysiaf.utils import rotations

from jwst_rogue_path_tool.utils import make_output_directory


def create_exposure_plots(observation, ra, dec, output_directory=None, **kwargs):
    """Generate exposure level plots

    Parameters
    ----------
    ra : float
        Right Ascension in degrees

    dec : float
        Declination in degrees

    **kwarg : dict
        Arbitrary keyword arguments
    """

    plt.rcParams["figure.figsize"] = (20, 15)

    inner_radius = kwargs.get("inner_radius", 8.0)
    outer_radius = kwargs.get("outer_radius", 12.0)
    ncols = kwargs.get("ncols", 4)

    wedge_length = inner_radius - 1
    exposure_frames = observation["exposure_frames"]
    exposure_frames_data = exposure_frames.data

    nrows = len(exposure_frames_data) // ncols + (len(exposure_frames_data) % ncols > 0)

    catalog = exposure_frames.catalog
    plotting_catalog = locate_targets_in_annulus(
        catalog, ra, dec, inner_radius, outer_radius
    )

    obs_id = observation["visit"]["observation"].values[0]
    program = observation["nircam_templates"]["program"].values[0]

    for n, exp_num in enumerate(exposure_frames_data):
        angle_start = exposure_frames.valid_starts_angles[exp_num]
        angle_end = exposure_frames.valid_ends_angles[exp_num]

        ax = plt.subplot(nrows, ncols, n + 1)
        ax.set_xlabel("RA [Degrees]")
        ax.set_ylabel("DEC [Degrees]")
        ax.set_title("Observation {}, Exposure: {}".format(obs_id, exp_num))
        ax.scatter(ra, dec, marker="X", c="red")
        ax.scatter(
            plotting_catalog["ra"],
            plotting_catalog["dec"],
            c="deeppink",
        )
        ax.axis("equal")
        ax.invert_xaxis()

        if angle_start is None and angle_end is None:
            ax.annotate("NO VALID ANGLES", xy=(ra, dec))
        else:
            for angles in zip(angle_start, angle_end):
                min_theta = (
                    90.0 - angles[0]
                )  # Origin of matplotlib is offset by 90 degrees
                max_theta = (
                    90.0 - angles[1]
                )  # Origin of matplotlib is offset by 90 degrees
                w = Wedge(
                    (ra, dec),
                    wedge_length,
                    max_theta,
                    min_theta,
                    fill=False,
                    color="darkseagreen",
                    joinstyle="round",
                )
                ax.add_artist(w)

            for angle in np.concatenate([angle_start, angle_end]):
                exposure_frames.calculate_attitude(angle)
                sus_region_patches = get_susceptibility_region_patch(
                    exposure_frames, exp_num
                )
                for patch in sus_region_patches:
                    ax.add_patch(patch)

    plt.tight_layout()

    if output_directory:
        filename = f"exposure_plot_program_{program}_obs_id_{obs_id}.png"
        full_path = output_directory / str(program) / "exposure_level"
        make_output_directory(full_path)
        print(f"WRITING FIGURE TO {full_path / filename}")
        plt.savefig(full_path / filename)
    else:
        plt.show()

    plt.close()


def create_observation_plot(observation, ra, dec, output_directory=None, **kwargs):
    """Plot that describe all valid angles at the observation level.
    The observation level plot is a single plot of all valid angles
    from a set of exposures.

    Parameters
    ----------
    ra : float
        Right Ascension in degrees

    dec : float
        Declination in degrees

    **kwarg : dict
        Arbitrary keyword arguments
    """

    plt.rcParams["figure.figsize"] = (10, 10)

    inner_radius = kwargs.get("inner_radius", 8.0)
    outer_radius = kwargs.get("outer_radius", 12.0)

    wedge_length = inner_radius - 1
    exposure_frames = observation["exposure_frames"]

    observation_number = observation["nircam_templates"]["observation"].values[0]
    program = observation["nircam_templates"]["program"].values[0]

    catalog = exposure_frames.catalog
    plotting_catalog = locate_targets_in_annulus(
        catalog, ra, dec, inner_radius, outer_radius
    )

    ax = plt.subplot()
    ax.set_xlabel("RA [Degrees]")
    ax.set_ylabel("DEC [Degrees]")
    ax.set_title(f"Program {program} Observation {observation_number}")
    ax.scatter(ra, dec, marker="X", c="red")
    ax.scatter(plotting_catalog["ra"], plotting_catalog["dec"], c="deeppink")

    ax.axis("equal")
    ax.invert_xaxis()

    all_starting_angles = observation["valid_starts_angles"]
    all_ending_angles = observation["valid_ends_angles"]

    if all_starting_angles.size == 0 and all_ending_angles.size == 0:
        ax.annotate("NO VALID ANGLES", xy=(ra, dec))
    else:
        for angles in zip(all_starting_angles, all_ending_angles):
            min_theta = 90.0 - angles[0]  # Origin of matplotlib is offset by 90 degrees
            max_theta = 90.0 - angles[1]  # Origin of matplotlib is offset by 90 degrees
            w = Wedge(
                (ra, dec),
                wedge_length,
                max_theta,
                min_theta,
                fill=False,
                color="darkseagreen",
                joinstyle="round",
            )
            ax.add_artist(w)

    plt.tight_layout()
    if output_directory:
        filename = f"observation_plot_program_{program}_obs_id_{observation_number}.png"
        full_path = output_directory / str(program) / "observation_level"
        make_output_directory(full_path)
        print(f"WRITING FIGURE TO {full_path / filename}")
        plt.savefig(full_path / filename)
    else:
        plt.show()

    plt.close()


def get_susceptibility_region_patch(exposure_frames, exposure_id):
    """Obtain data for susceptibility region and generate plottable
    patch.

    Parameters
    ----------
    exposure_frames : jwst_rogue_path_tool.detect_claws.ExposureFrames
        ExposureFrame object associated with observation.
    """
    patches = []

    region = exposure_frames.susceptibility_region

    for key in region:
        module = region[key]
        attitude = exposure_frames.attitude

        v2 = 3600 * module.V2V3path.vertices.T[0]  # degrees --> arcseconds
        v3 = 3600 * module.V2V3path.vertices.T[1]  # degrees --> arcseconds

        ra_rads, dec_rads = rotations.tel_to_sky(
            attitude, v2, v3
        )  # returns ra and dec in radians
        ra_deg, dec_deg = (
            ra_rads * 180.0 / np.pi,
            dec_rads * 180.0 / np.pi,
        )  # convert to degrees

        ra_dec_path = Path(np.array([ra_deg, dec_deg]).T, module.V2V3path.codes)
        ra_dec_patch = PathPatch(ra_dec_path, lw=2, alpha=0.1)
        patches.append(ra_dec_patch)

    return patches


def locate_targets_in_annulus(catalog, ra, dec, inner_radius, outer_radius):
    """Calculate the targets from a catalog that fall within inner and outer radii.

    Parameters
    ----------
    catalog : pandas.core.frame.DataFrame
        DataFrame of star positions and magnitudes from 2MASS

    ra : float
        Right Ascension in degrees

    dec : float
        Declination in degrees

    inner_radius : float
        Inner radius of annulus

    outer_radius : float
        Outer area of annulus
    """

    # Set coordinates for target and catalog
    target_coordinates = SkyCoord(ra * u.deg, dec * u.deg, frame="icrs")
    catalog_coordinates = SkyCoord(
        catalog["ra"].values * u.deg,
        catalog["dec"].values * u.deg,
        frame="icrs",
    )

    # Calculate separation from target to all targets in catalog
    separation = target_coordinates.separation(catalog_coordinates)
    mask = (separation.deg < outer_radius) & (separation.deg > inner_radius)

    # Retrieve all targets in masked region above.
    plotting_catalog = catalog[mask]

    return plotting_catalog


def plot_fixed_angle_regions(observation, angle, output_directory=None):
    """Plot the susceptibility region and targets in and around it.

    Parameters
    ----------
    observation : dictionary
        Dictionary of a single observation dataset

    angle : float
        Angle of attitude.

    output_directory : str
        Directory to save figure to.
    """
    program = observation["nircam_templates"]["program"].values[0]
    observation_number = observation["nircam_templates"]["observation"].values[0]
    susceptibility_region = observation["exposure_frames"].susceptibility_region
    number_of_modules = len(susceptibility_region)
    modules_name = observation["nircam_templates"]["modules"].values[0]

    plt.set_cmap("magma")
    fig, ax = plt.subplots(number_of_modules, figsize=(15, 15))

    # Hack for the loop below to work with a program that contains
    # a single module (A or B) or both modules (A and B).
    try:
        len(ax)
    except TypeError:
        ax = [ax]

    for ax, module in zip(ax, susceptibility_region):
        centroid = susceptibility_region[module].centroid
        averages = observation[f"averages_{module}"][angle]

        avg_v2 = averages[f"avg_v2_{module}"]
        avg_v3 = averages[f"avg_v3_{module}"]
        avg_intensity = averages[f"avg_intensity_{module}"]

        im = ax.scatter(avg_v2, avg_v3, c=avg_intensity)
        fig.colorbar(im, ax=ax, label="Intensity")

        # Make box around centroid of centroid of susceptibility region.
        ax.set_xlim([centroid[0] - 3, centroid[0] + 3])
        ax.set_ylim([centroid[1] - 3, centroid[1] + 3])

        ax.set_xlabel("V2")
        ax.set_ylabel("V3")
        ax.title.set_text(
            f"Program: {program} Observation: {observation_number} Angle: {angle} Module: {module}"
        )

        patch = PathPatch(susceptibility_region[module].V2V3path, alpha=0.1)
        ax.add_patch(patch)

    if output_directory:
        full_path = output_directory / str(program) / "susceptibility_regions"
        filename = f"{program}_{observation_number}_{modules_name}_{angle}.png"
        make_output_directory(full_path)
        print(f"WRITING FIGURE TO {full_path / filename}")
        plt.savefig(full_path / filename)
    else:
        plt.show(fig)

    plt.close(fig)


def create_v3pa_vs_flux_plot(observation, output_directory=None, fontsize=15):
    """Create plot of V3 position angles and flux for a single observation.

    Parameters
    ----------
    observation : dictionary
        Dictionary of data associated with a single observation.

    output_dictionary : str
        Directory to save figure to

    fontsize : int
        Fontsize for plot
    """
    observation_number = observation["nircam_templates"]["observation"].values[0]
    program_id = observation["visit"]["program"].values[0]
    susceptibility_regions = observation["exposure_frames"].susceptibility_region
    modules = susceptibility_regions.keys()
    filters = observation["filters"]
    pupils = observation["pupils"]

    flux = observation["flux"]["dn_pix_ks"]
    flux_boolean = observation["flux_boolean"]

    colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]

    fig, axes = plt.subplots(
        len(filters),
        len(modules),
        figsize=(15, 10),
        sharex=True,
        sharey=True,
        squeeze=False,
    )

    # Assign module name to columns of plots.
    column_names = [f"Module: {module}" for module in modules]
    for ax, col in zip(axes[0], column_names):
        ax.set_title(col, fontsize=fontsize)

    fig.suptitle(
        f"Program: {program_id} Observation: {observation_number}", fontsize=fontsize
    )

    for mod, module in enumerate(modules):
        for fltr, (filter, pupil) in enumerate(zip(filters, pupils)):
            flux_key = f"dn_pix_ks_{pupil}+{filter}_{module}"
            flux_values = flux[flux_key]
            axes[fltr, mod].plot(flux_values)
            above_threshold = np.copy(flux_values)

            for color_idx, key in enumerate(flux_boolean[f"{filter}_{module}"].keys()):
                stats_function, lam_threshold, bkg_threshold = key.split("_")
                mask = flux_boolean[f"flux_boolean_{stats_function}_{module}"]
                for other_module in modules:
                    if other_module != module:
                        mask &= flux_boolean[f"flux_boolean_{stats_function}_{other_module}"]
                above_threshold[mask] = np.nan
                label_str = f"{eval(bkg_threshold):.1f} x {stats_function} = {eval(lam_threshold):.1f} DN/pix/ks"
                axes[fltr, mod].plot(above_threshold, c=colors[color_idx + 1])
                axes[fltr, mod].axhline(
                    eval(lam_threshold),
                    c=colors[color_idx + 1],
                    ls="--",
                    label=label_str,
                )

            axes[fltr, mod].set_yscale("log")
            axes[fltr, mod].set_xlabel("V3PA", fontsize=fontsize)
            axes[fltr, mod].set_ylabel(f"DN/pix/ks ({pupil if pupil != 'CLEAR' else filter})", fontsize=fontsize)
            axes[fltr, mod].set_ylim(0.005, 500)
            axes[fltr, mod].legend(
                loc="lower right", fontsize=fontsize - (fontsize / 4)
            )

    axes[0, 0].set_xlim(0, 360)
    fig.tight_layout()

    if output_directory:
        if len(modules) > 1:
            filename = f"v3pa_vs_flux_{program_id}_{observation_number}_ALL.png"
        else:
            filename = f"v3pa_vs_flux_{program_id}_{observation_number}_{module}.png"

        full_path = output_directory / str(program_id) / "v3pa_vs_flux"
        make_output_directory(full_path)
        print(f"WRITING FIGURE TO {full_path / filename}")
        plt.savefig(full_path / filename)
    else:
        plt.show(fig)

    plt.close(fig)



"""
This module contains the FixedAngle class which performs the analysis for
JWST Rogue Path Tool.

This routine accepts an observation from the aptProgram class

Authors
-------
    - Mario Gennaro
    - Mees Fix
"""

import numpy as np

from jwst_rogue_path_tool.constants import (
    CATALOG_BANDPASS,
    NIRCAM_ZEROPOINTS,
    ZEROPOINT,
)

from jwst_rogue_path_tool.plotting import plot_fixed_angle_regions
from jwst_rogue_path_tool.utils import absolute_magnitude, get_pupil_from_filter

np.seterr(divide="ignore")  # For np.log10 divide by 0


class fixedAngle:
    """Analysis for a single fixed V3 position angle. For a specific observation and V3PA
    the magnitudes/fluxes of stars in and around the susceptibility region are calculated.

    Parameters
    ----------
    observation : dictionary
        Observation entry from aptProgram class

    angle : float
        V3 position angle
    """

    def __init__(self, observation, angle):
        self.observation = observation
        self.angle = angle
        self.catalog = self.observation["exposure_frames"].catalog
        self.catalog_name = self.observation["exposure_frames"].catalog_name
        self.bands = CATALOG_BANDPASS[self.catalog_name]

        self.exposure_frame_table = self.observation[
            "exposure_frames"
        ].exposure_frame_table

        self.susceptibility_region = self.observation[
            "exposure_frames"
        ].susceptibility_region

        self.filter_modules_combos = self.exposure_frame_table.groupby(
            ["filter_short", "modules"]
        ).size()

        self.total_exposure_duration_table = self.observation[
            "exposure_frames"
        ].total_exposure_duration_table

        self.filters = self.total_exposure_duration_table["filters"]
        self.pupils = self.total_exposure_duration_table["pupils"]
        self.calculate_absolute_magnitude()
        self.get_total_magnitudes()
        self.get_total_counts()

    def calculate_absolute_magnitude(self):
        """Calculate the absolute magnitude of the intensities averaged over exposures
        for the specific V3 position angle.
        """
        for module in self.susceptibility_region:
            for band in self.bands:
                average_intensity = self.observation[f"averages_{module}"][self.angle][
                    f"avg_intensity_{module}"
                ]
                abs_magnitude = (
                    self.catalog[band]
                    + absolute_magnitude(average_intensity)
                    + ZEROPOINT
                )

                # Test and see if this can be removed, using np.inf should be fine.
                abs_magnitude.replace(-np.inf, np.nan, inplace=True)

                self.catalog[f"abs_mag_{band}_{module}"] = abs_magnitude

    def get_empirical_zero_points(self, module, pupil, filter):
        """Get NRC emperical zero point points for pupil/filter combination for module A or B.

        Parameters
        ----------
        module : str
            NRC module name ("A" or "B")

        pupil : str
            NRC pupil name

        filter : str
            NRC filter name
        """
        if module == "A":
            return NIRCAM_ZEROPOINTS["zeropoints_A"][f"{pupil}+{filter}"]
        elif module == "B":
            return NIRCAM_ZEROPOINTS["zeropoints_B"][f"{pupil}+{filter}"]
        else:
            assert ValueError(f"module must be value 'A' or 'B' not {module}")

    def get_ground_band(self, pupil, filter, catalog="2MASS"):
        """Get band name from catalog corresponding to JWST NRC pupil/filter combination

        Parameters
        ----------
        pupil : str
            NRC pupil name

        filter : str
            NRC filter name

        catalog : str
            Name of source catalog (default: "2MASS")
        """
        if catalog == "2MASS":
            return NIRCAM_ZEROPOINTS["match2MASS"][f"{pupil}+{filter}"]
        else:
            assert ValueError(f"Catalog must be value '2MASS' not {catalog}")

    def get_total_counts(self):
        """Calculate the total number of counts based on the calculated magnitudes and
        exposure durations.
        """
        self.total_counts = {}
        for module in self.susceptibility_region:
            for _, row in self.total_exposure_duration_table.iterrows():
                filter = row["filters"]
                pupil = row["pupils"]

                ground_band = self.get_ground_band(pupil, filter)
                # Obtain emperical (Vega Mag) zeropoint based on filter/pupil combo
                emperical_zeropoint = self.get_empirical_zero_points(
                    module, pupil, filter
                )

                # Get summed abs magnitude for entire band.
                summed_abs_mag = self.total_magnitudes[
                    f"total_mag_{ground_band}_{module}"
                ]

                total_exposure_duration = self.total_exposure_duration_table.loc[
                    (self.total_exposure_duration_table["pupils"] == pupil)
                    & (self.total_exposure_duration_table["filters"] == filter)
                ]["photon_collecting_duration"].values[0]
                tot_cnts = (
                    10 ** ((emperical_zeropoint - summed_abs_mag) / 2.5)
                    * total_exposure_duration
                )

                self.total_counts[f"total_counts_{pupil}+{filter}_{module}"] = tot_cnts

    def get_total_magnitudes(self):
        """Calculate the total (summed) magnitudes for each band in catalog."""
        self.total_magnitudes = {}
        for module in self.susceptibility_region:
            for band in self.bands:
                column_name = f"abs_mag_{band}_{module}"
                if column_name in self.catalog.columns:
                    summed_magnitude = np.nansum(
                        np.power(10, -0.4 * self.catalog[column_name].values)
                    )
                    total_mag = absolute_magnitude(summed_magnitude)
                else:
                    raise Exception(f"{column_name} not available in catalog!")
                self.total_magnitudes[f"total_mag_{band}_{module}"] = total_mag

    def plot_regions(self):
        """Convenience method for plotting susceptibility region for a single angle
        in V2/V3 space with maginitude intensity."""
        plot_fixed_angle_regions(self.observation, self.angle)



"""This module contains all of the routines for parsing a JWST APT JSON output files

Authors
-------
    - Mario Gennaro
    - Mees Fix

Use
---
    Routines in this module can be imported as follows:

    >>> from jwst_rogue_path_tool.program_data_parser import aptJsonFile
    >>> filename = "/path/to/apt_json_file.json"
    >>> json = aptJsonFile(filename)
"""

from astropy.table import Table
import json
import pandas as pd


class aptJsonFile:
    """Read and parse JSON file generated from APT"""

    def __init__(self, json_file):
        """
        Parameters
        ----------
        json_file : str
            Name of APT JSON file
        """
        self.json_file = json_file
        self.data = self.read_json_file()
        self.tablenames = list(self.data.keys())

    def build_dataframe(self, tablename, show_table=False):
        """Create pandas dataframe from parsed APT JSON file

        Parameters
        ----------
        tablename : str
            Name of table to generate pandas dataframe from.

        show_table : bool
            Show table in web browser.
        """
        df = pd.DataFrame(self.data[tablename])
        df = df.apply(pd.to_numeric, errors="coerce").fillna(df)

        # Show table in web browser
        if show_table:
            t = Table.from_pandas(df)
            t.show_in_browser()

        return df

    def read_json_file(self):
        """Read JSON file exported by APT"""

        with open(self.json_file, "r") as file:
            data = json.load(file)

        return data



"""Module that writes out all outputs by default for APT SQL files.

Authors
-------
    - Mees Fix
"""

import warnings

import argparse
import pathlib

from jwst_rogue_path_tool.constants import PROJECT_DIRNAME
from jwst_rogue_path_tool.detect_claws import aptProgram


def process_programs(apt_programs, output_directory):
    """Process and write output from program."""
    for apt_program in apt_programs:
        print(
            f"Processing program file {apt_program} to output directory {output_directory}"
        )
        try:
            program = aptProgram(apt_program, angular_step=1.0)
            program.run()
            supported_observations = program.observations.supported_observations.keys()

            for obs_number in supported_observations:
                observation = program.observations.data[obs_number]
                program.plot_exposures(observation, output_directory)
                program.plot_observation(observation, output_directory)
                program.plot_v3pa_vs_flux(observation, output_directory)
                program.make_report(observation, output_directory)

            del program
        except Exception as e:
            warnings.warn(f"EXCEPTION FOR PROGRAM {apt_program}: {e}")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--filename", type=str, help="Filename or filename search pattern"
    )
    parser.add_argument("--output_dir", type=str, help="Directory to write output")
    args = parser.parse_args()

    if not args.output_dir:
        raise Exception("No output directory specified!")
    else:
        output_dir = pathlib.Path(args.output_dir)
        if not output_dir.is_dir():
            raise Exception(f"{args.output_dir} is not a directory!")

    files_path = pathlib.Path(args.filename)
    if files_path.is_file():
        process_programs([files_path], output_dir)
    else:
        file_path = files_path.parent
        file_pattern = files_path.name
        all_files = file_path.glob(file_pattern)
        print(all_files, file_pattern, file_path)
        process_programs(all_files, output_dir)



"""
Module to predict the presence of the stray-light artifacts know as claws
in NIRCam Imaging observations, specified in APT.

Authors
-------
    - Mario Gennaro
"""

from astroquery.simbad import Simbad
import pysiaf
from pysiaf.utils import rotations
from jwst_backgrounds import jbt

from .apt_sql import Sqlfile

import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits

import matplotlib.pyplot as plt
import matplotlib.patches as patches
from matplotlib.path import Path

import numpy as np
import pandas as pd
from scipy.ndimage import gaussian_filter


class apt_program:
    """
    Class that handles the APT-program-level information.
    It can configure "observation" objects based on the desired observation ids,
    and can cal the observation.check_multiple_angles method to perform
    a check of stars in the susceptibility region for all the exposures of a
    given observation and for multiple observations of a given program
    """

    def __init__(self, sqlfile, instrument="NIRCAM"):
        """
        Parameters
        ----------
        sqlfile : str
            Path to an APT-exported sql file

        catargs : dictionary
            parameters to be passed to the get_catalog method of an observation object
        """

        sql = Sqlfile(sqlfile)

        self.exptable = sql.table("exposures").to_pandas()
        try:
            self.targtable = sql.table("fixed_target").to_pandas()
        except:
            print("**** This is a moving target program, exiting ****")
            assert False
        self.nrctemptable = sql.table("nircam_templates").to_pandas()
        self.nestable = sql.table("nircam_exposure_specification").to_pandas()
        self.visittable = sql.table("visit").to_pandas()

        self.instrument = instrument
        self.PID = self.exptable.iloc[0]["program"]
        self.siaf = pysiaf.Siaf(self.instrument)
        self.observations = []

    def configure_observation(self, obsid, catargs, smallregion=False, nes=None):
        """
        Create an observation object

        Parameters
        ----------
        obsid: index of a Pandas dataframe
            Represents the observation ID of interest

        catargs: dictionary
            parameters to be passed to the get_catalog method of an observation object

        smallregion: boolean
            if True restricts the search to a smaller susceptibility region
            in the rogue path

        nes: integer
            specifies which exposure_spec_order_number to use. If None, the code will
            select the exposure_spec_order_number corresponding to the highest number
            of expected counts (based on photon collecting time and zero point value)

        Returns
        ----------
        An observation object. Returns None if no exposures are present with the input
        obsid in the APT-exported sql file
        """

        print("Adding obsid:", obsid)

        if self.instrument == "NIRCAM":
            prefix = "NRC"

        obsskip = False
        nrcused = False
        nrcparallel = False
        BM = self.visittable["observation"] == obsid
        fvt = self.visittable[BM].iloc[0]
        if (fvt["template"] != "NIRCam Imaging") & (
            fvt["template"] != "NIRCam Wide Field Slitless Spectroscopy"
        ):
            if "template_coord_parallel_1" in self.visittable.columns:
                if (fvt["template_coord_parallel_1"] != "NIRCam Imaging") & (
                    fvt["template_coord_parallel_1"]
                    != "NIRCam Wide Field Slitless Spectroscopy"
                ):
                    obsskip = True
                    if "NIRC" in fvt["template"]:
                        template_name = fvt["template"]
                        nrcused = True
                    elif "NIRC" in fvt["template_coord_parallel_1"]:
                        template_name = fvt["template_coord_parallel_1"]
                        nrcused = True
                else:
                    nrcparallel = True
                    nrcused = True

            else:
                obsskip = True
                if "NIRC" in fvt["template"]:
                    template_name = fvt["template"]
                    nrcused = True
        else:
            nrcused = True

        if obsskip:
            if nrcused:
                print(
                    "**** The {} template is not supported, obsid {} will not be added ****".format(
                        template_name, obsid
                    )
                )
            else:
                print(
                    "**** obsid {} does not use NIRCam and will not be added ****".format(
                        obsid
                    )
                )
            return None
        else:
            if nrcparallel:
                print("**** NIRCam imaging used in parallel ****")

        BM = (
            (self.exptable["observation"] == obsid)
            & (self.exptable["pointing_type"] == "SCIENCE")
            & (self.exptable["AperName"].str.contains(prefix))
        )

        if np.sum(BM) == 0:
            print(
                "No {} expsoures in this program for obsid {}".format(
                    self.instrument, obsid
                )
            )
            return None
        print(
            "Total number of {} exposures in this observation: {}".format(
                self.instrument, np.sum(BM)
            )
        )
        exptable_obs = self.exptable[BM]

        BM = self.nestable["observation"] == obsid
        print(
            "Total number of {} exposure specifications: {}".format(
                self.instrument, np.sum(BM)
            )
        )
        nestable_obs = self.nestable[BM]

        BM = self.visittable["observation"] == obsid
        print("Total number of visits: {}".format(np.sum(BM)))
        visittable_obs = self.visittable[BM]

        BM = self.nrctemptable["observation"] == obsid
        modules = self.nrctemptable[BM].iloc[0]["modules"]

        ##### This part must go. Refactor the check one angle to check only for a single exposure specification

        #       nes_ids = nestable['order_number']
        #       pcds = np.empty(len(nes_ids))
        #       zps = np.empty(len(nes_ids))
        #       ps = np.empty(len(nes_ids),dtype=object)
        #       fs = np.empty(len(nes_ids),dtype=object)
        #
        #       zpc = zero_point_calc()
        #       print('{:4} {:8} {:8} {:6} {:6} {:6}'.format('ID','PUPIL','FILTER','T_exp','   ZP','  K'))
        #
        #       for i,nes_id in enumerate(nes_ids):
        #           BM = exptable_obs['exposure_spec_order_number'] == nes_id
        #           pcds[i] = exptable_obs.loc[BM,'photon_collecting_duration'].values.astype(np.float_).sum()
        #
        #           BM = nestable['order_number'] == nes_id
        #           filtershort = nestable.loc[BM,'filter_short'].values[0]
        #           if '_' in filtershort:
        #               splt = filtershort.split('_')
        #               pupilshort = splt[0]
        #               filtershort = splt[1]
        #           else:
        #               pupilshort = 'CLEAR'
        #           zps[i] = zpc.get_avg_zp(pupilshort,filtershort)
        #           ps[i],fs[i] = pupilshort, filtershort
        #           print('{:4} {:8} {:8} {:6.0f} {:6.2f} {:.2e}'.format(i,ps[i],fs[i],pcds[i],zps[i],pcds[i]*10**(zps[i]/2.5)))
        #
        #       if nes is None:
        #           cts_estimate = pcds*10**(zps/2.5)
        #           ines = np.argmax(cts_estimate)
        #           idxnes = nes_ids.index[ines]
        #           nes = nes_ids[idxnes]
        #           print('Exposure specification to use:{}, (idx={}); using PUPILSHORT={}, FILTERSHORT={}'.format(nes,ines,ps[ines],fs[ines]))
        #       else:
        #           print('Forcing use of exposure specification:{}, (idx={}); using PUPILSHORT={}, FILTERSHORT={}'.format(nes,nes+1,ps[nes],fs[nes]))
        #
        #
        #
        #       BM = exptable_obs['exposure_spec_order_number'] == nes
        #       exptable_obs = exptable_obs[BM]
        #
        #
        #       BM = nestable['order_number'] == nes
        #       nestable_obs = nestable[BM]

        targetrowidx = self.targtable["target_id"] == exptable_obs.iloc[0]["target_id"]
        target_ra = self.targtable.loc[targetrowidx, "ra_computed"].values[0]
        target_dec = self.targtable.loc[targetrowidx, "dec_computed"].values[0]

        return observation(
            exptable_obs,
            nestable_obs,
            visittable_obs,
            target_ra,
            target_dec,
            modules,
            catargs,
            self.siaf,
            smallregion=smallregion,
        )

    def add_observations(
        self,
        obsids=None,
        catargs={
            "inner_rad": 8.0,
            "outer_rad": 12.0,
            "sourcecat": "SIMBAD",
            "band": "K",
            "maxmag": 4.0,
            "simbad_timeout": 200,
            "verbose": True,
        },
        smallregion=False,
    ):
        """
        Configure multiple observations and append them to the self.observations list

        Parameters
        ----------
        obsid: list of Pandas indexes
            IDs of the observations to add. If None, all the opbservations in the prgram
            will be added

        catargs: dictionary
            parameters to be passed to the get_catalog method of an observation object
        """

        if obsids is None:
            obsids = self.exptable["observation"].unique()
        for obsid in obsids:
            added = False
            for obs in self.observations:
                if obsid == obs.obsid:
                    print("Obsid {} already added".format(obsid))
                    added = True
                    break
            if added == False:
                obshere = self.configure_observation(
                    obsid, catargs, smallregion=smallregion
                )
                if obshere is not None:
                    self.observations.append(obshere)

    def check_observations(self, obsids=None, angstep=0.5, RP_padding=0):
        """
        Convenience function to check multiple observations for stars in the
        susceptibility region

        Parameters
        ----------
        obsid: list of Pandas indexes
            IDs of the observations to check. If None, all the opbservations in the prgram
            will be added

        angstep: float
            The resolution at which to scan the whole (0,360) range of PAs, in degrees

        RP_padding: float
            Extra padding around the susceptibility region (stars outside the nominal
            SR, but within RP_padding are flagged as "inside")
        """

        if obsids is None:
            obsids = self.exptable["observation"].unique()
        for obsid in obsids:
            for obs in self.observations:
                if obs.obsid == obsid:
                    print("Checking obsid:", obsid)
                    obs.check_multiple_angles(angstep, RP_padding=0.0)


class observation:
    """
    Class that handles an indivdual observation.
    It contains info on each exposure pointing configuration,
    retrieves a catalog within a certain annulus, and can check for stars
    in the susceptibility region
    """

    def __init__(
        self,
        exptable_obs,
        nestable_obs,
        visittable_obs,
        target_ra,
        target_dec,
        modules,
        catargs,
        siaf,
        smallregion=False,
    ):
        """
        Parameters
        ----------
        exptable_obs: Pandas dataframe
            contains one row from the exposure table, per each exposure within the observation

        nestable_obs: Pandas dataframe
            contains one row from nircam_exposures_specification table for each exposure specification
             selected for this observation

        visittable_obs: Pandas dataframe
            contains one row from the visit table for each visit associated to this observation


        target_ra,target_dec: floats
            Coorindtaes of the target of the obsrvation in decimal degreess

        modules: list of strings
            name of the nircam modules configuration for the exposures within this observation
            (note that in NIRCam imaging, all the exposures of an observation must have the same configuration)

        catargs: dictionary
            parameters to be passed to the get_catalog method

        siaf: pysiaf.Siaf instance
            object containing the apertures info for NIRCam
        """

        self.exptable_obs = exptable_obs
        self.nestable_obs = nestable_obs
        self.obsid = self.exptable_obs.iloc[0]["observation"]
        self.program = self.exptable_obs.iloc[0]["program"]
        self.target_ra = target_ra
        self.target_dec = target_dec
        self.modules = modules
        self.catargs = catargs
        self.smallregion = smallregion
        self.SRlist, self.SRnames = self.get_SRlist()
        self.efs = self.get_exposure_frames(siaf)
        self.catdf = self.get_catalog()

    def get_SRlist(self):
        """
        Parameters
        ----------
        None

        Returns
        ----------
        SRlist: list of matplotlib.Path objects
            The list of susceptibility regions corresponding to the module used in this observation
        """

        if (self.modules == "ALL") | (self.modules == "BOTH"):
            SRlist = [
                sus_reg(module="A", small=self.smallregion),
                sus_reg(module="B", small=self.smallregion),
            ]
            SRnames = ["A", "B"]
        else:
            if self.modules[0] == "A":
                SRlist = [sus_reg(module="A", small=self.smallregion)]
                SRnames = ["A"]
            elif self.modules[0] == "B":
                SRlist = [sus_reg(module="B", small=self.smallregion)]
                SRnames = ["B"]

        return SRlist, SRnames

    def get_exposure_frames(self, siaf):
        """
        Parameters
        ----------
        siaf:
            a pysiaf.Siaf instance

        Returns
        ----------
        efs: list of exposure_frame objects
            The list of objects containing pointing info for each exposure within this observation
        """

        efs = []
        for i, row in self.exptable_obs.iterrows():
            BM = (self.nestable_obs["visit"] == row["visit"]) & (
                self.nestable_obs["order_number"] == row["exposure_spec_order_number"]
            )
            efs.append(exposure_frame(row, siaf, self.nestable_obs.loc[BM]))
        return efs

    def get_catalog(self):
        """
        Parameters
        ----------
        None

        Returns
        ----------
        df: Pandas dataframe
            dataframe containing the coordinates of stars within an annulus centered on the
            observation target. The catalog characterisitics are based on the catargs dictionary

        """

        # Get the maximum relative offset between exposures and pad the catalog search radius
        ra_cen_sorted = np.sort(
            self.exptable_obs["ra_center_rotation"].values.astype(np.float_)
        )
        dec_cen_sorted = np.sort(
            self.exptable_obs["dec_center_rotation"].values.astype(np.float_)
        )

        max_ra_diff = np.abs(ra_cen_sorted[-1] - ra_cen_sorted[0])
        max_dec_diff = np.abs(dec_cen_sorted[-1] - dec_cen_sorted[0])

        max_delta = np.sqrt(np.sum(np.square([max_ra_diff, max_dec_diff])))

        inner_rad = self.catargs["inner_rad"] - max_delta
        outer_rad = self.catargs["outer_rad"] + max_delta

        if self.catargs["verbose"]:
            print("Adopted inner and outer radius", inner_rad, outer_rad)

        # Retrieve a catalog
        if self.catargs["sourcecat"] == "SIMBAD":
            df_in = querysimbad(
                self.target_ra,
                self.target_dec,
                rad=inner_rad,
                band=self.catargs["band"],
                maxmag=self.catargs["maxmag"],
                simbad_timeout=self.catargs["simbad_timeout"],
            ).to_pandas()
            df_out = querysimbad(
                self.target_ra,
                self.target_dec,
                rad=outer_rad,
                band=self.catargs["band"],
                maxmag=self.catargs["maxmag"],
                simbad_timeout=self.catargs["simbad_timeout"],
            ).to_pandas()

            df = pd.concat([df_in, df_out]).drop_duplicates(keep=False)
            for i, row in df.iterrows():
                coord = SkyCoord(row["RA"], row["DEC"], unit=(u.hourangle, u.deg))
                df.loc[i, "RAdeg"] = coord.ra.deg
                df.loc[i, "DECdeg"] = coord.dec.deg

        if self.catargs["sourcecat"] == "2MASS":
            df = pd.read_csv(self.catargs["2MASS_filename"])
            BM = df[self.catargs["band"]] < self.catargs["maxmag"]
            df = df[BM]

            c1 = SkyCoord(self.target_ra * u.deg, self.target_dec * u.deg, frame="icrs")
            c2 = SkyCoord(df["ra"] * u.deg, df["dec"] * u.deg, frame="icrs")
            sep = c1.separation(c2)
            BM = (sep.deg < outer_rad) & (sep.deg > inner_rad)

            df = df[BM]
            df.rename(columns={"ra": "RAdeg", "dec": "DECdeg"}, inplace=True)

        return df

    def check_multiple_angles(self, angstep, filtershort=None, RP_padding=0.0):
        """
        Convenience method to check multiple angles at once, for all
        exposures within this observation

        Parameters
        ----------
        angstep: float
            The resolution at which to scan the whole (0,360) range of PAs, in degrees

        filtershort: string
            name of the filter in use, will be used to downselect the exposures.
            if None will be defaulted to the first filter in the nircam_expsoures_specification
            table

        RP_padding: float
            Extra padding around the susceptibility region (stars outside the nominal
            SR, but within RP_padding are flagged as "inside")
        """

        if filtershort is None:
            filtershort = self.nestable_obs["filter_short"].values[0]

        efs_here = [
            ef
            for ef in self.efs
            if ef.nestable_row["filter_short"].values[0] == filtershort
        ]

        self.angstep = angstep
        self.RP_padding = RP_padding
        self.attitudes = np.arange(0.0, 360.0, angstep)
        self.IN = np.empty(
            [len(self.catdf), self.attitudes.size, len(self.SRlist), len(efs_here)],
            dtype=np.bool_,
        )
        self.V2 = np.empty([len(self.catdf), self.attitudes.size, len(efs_here)])
        self.V3 = np.empty([len(self.catdf), self.attitudes.size, len(efs_here)])
        self.good_angles = np.zeros(
            [self.attitudes.size, len(efs_here)], dtype=np.bool_
        )

        for i, att in enumerate(self.attitudes):
            IN_one, V2_one, V3_one, check_one = self.check_one_angle(
                att, filtershort, RP_padding=self.RP_padding
            )
            self.V2[:, i, :], self.V3[:, i, :] = V2_one, V3_one
            self.IN[:, i, :, :] = IN_one
            self.good_angles[i, :] = check_one

        V3PA_validranges_starts = []
        V3PA_validranges_ends = []

        for k in range(len(efs_here)):
            change = np.where(self.good_angles[:-1, k] != self.good_angles[1:, k])[0]

            if change.size > 0:
                if self.good_angles[change[0], k]:
                    change = np.roll(change, 1)

                V3PA_validranges_starts.append(self.angstep * change[::2])
                V3PA_validranges_ends.append(self.angstep * change[1::2])
        else:
            V3PA_validranges_starts.append(None)
            V3PA_validranges_ends.append(None)

        self.good_angles_obs = np.all(self.good_angles, axis=1)

        V3PA_validranges_obs_starts = []
        V3PA_validranges_obs_ends = []

        change = np.where(self.good_angles_obs[:-1] != self.good_angles_obs[1:])[0]
        if change.size > 0:
            if self.good_angles_obs[change[0]]:
                change = np.roll(change, 1)

            V3PA_validranges_obs_starts = self.angstep * change[::2]
            V3PA_validranges_obs_ends = self.angstep * change[1::2]
        else:
            V3PA_validranges_obs_starts = None
            V3PA_validranges_obs_ends = None

        self.V3PA_validranges_starts = V3PA_validranges_starts
        self.V3PA_validranges_ends = V3PA_validranges_ends
        self.V3PA_validranges_obs_starts = V3PA_validranges_obs_starts
        self.V3PA_validranges_obs_ends = V3PA_validranges_obs_ends

    def check_one_angle(self, att, filtershort, RP_padding=0.0):
        """
        Method to check for the presence of stars in the susceptibility
        region at a fixed angle, for all exposures of a given observation

        Parameters
        ----------
        att: float
            position angle to check, in degrees

        filtershort: string
            name of the filter in use, will be used to downselect the exposures

        RP_padding: float
            Extra padding around the susceptibility region (stars outside the nominal
            SR, but within RP_padding are flagged as "inside")

        Returns
        ---------
        IN_one: numpy boolean array (catalog size x number of SR region x number of exposures in the observation)
            True values indicate stars that are in (one of) the SRs for one of the exposures
        V2_one, V3one: numpy array (catalog size x number of exposures in the observation)
            V2, V3 coordinates (jn deg) for a given stars and a given exposure
        check_one: numpy boolean array (number of exposures in the observation)
            True if any star is in either SRs for a given exposure
        """

        efs_here = [
            ef
            for ef in self.efs
            if ef.nestable_row["filter_short"].values[0] == filtershort
        ]

        IN_one = np.empty(
            [len(self.catdf), len(self.SRlist), len(efs_here)], dtype=np.bool_
        )
        V2_one = np.empty([len(self.catdf), len(efs_here)])
        V3_one = np.empty([len(self.catdf), len(efs_here)])
        check_one = np.zeros(len(efs_here), dtype=np.bool_)

        for k, ef in enumerate(efs_here):
            ef.define_attitude(att)
            V2_one[:, k], V3_one[:, k] = ef.V2V3_at_one_attitude(
                self.catdf["RAdeg"], self.catdf["DECdeg"]
            )

            for j, SR in enumerate(self.SRlist):
                IN_one[:, j, k] = SR.V2V3path.contains_points(
                    np.array([V2_one[:, k], V3_one[:, k]]).T, radius=RP_padding
                )
            if len(self.SRlist) > 1:
                if ~(np.any(IN_one[:, 0, k]) | np.any(IN_one[:, 1, k])):
                    check_one[k] = True
            else:
                if ~(np.any(IN_one[:, 0, k])):
                    check_one[k] = True

        return IN_one, V2_one, V3_one, check_one

    def plot_obs_field(self, bandpass):
        """
        Method to plot the stars in the catalog

        Parameters
        ----------
        bandpass: string
            bandpass to use for color-coding the stars
        """

        f, ax = plt.subplots(1, 1)

        ax.scatter(self.catdf["RAdeg"], self.catdf["DECdeg"], c=self.catdf[bandpass])
        for ef in self.efs:
            ax.scatter(ef.raRef, ef.decRef, marker="o", c="orange")
        ax.scatter(self.target_ra, self.target_dec, marker="X", c="red")

        ax.axis("equal")
        ax.set_xlabel("RA")
        ax.set_ylabel("Dec")
        ax.invert_xaxis()
        f.tight_layout()

    def DN_report(
        self,
        attitudes,
        RP_padding=0.0,
        draw_reports=True,
        background_params=[{"threshold": 0.1, "func": np.min}],
        save_report_dir=None,
        save_figures_dir=None,
        verbose=False,
        smooth=None,
    ):
        """
        Method to call fixed_angle multiple times and get an estimated DN/pix/s for each
        filter in this observation, as a function of v3PA

        Paramters
        ---------

        attitudes: numpy array
            values of the attitudes at which one wants to compute the DN

        background_params: list of dictionarie (can be None)
           the claw flux is compared to threshold*func(background)
           where func is np.mean/np.min/np.max/np.median
           (or other callable that returns some summary stats),
           for each of the items of the list

        """

        tmA = []
        tmB = []
        tcA = []
        tcB = []

        for i, att in enumerate(attitudes):
            rd = self.fixed_angle(
                att,
                RP_padding=RP_padding,
                draw_allexp=False,
                draw_summary=False,
                smooth=smooth,
            )
            if i == 0:
                tot_exp_dur = rd["tot_exp_dur"]
                filtershort_all = rd["filtershort_all"]
                filternames = rd["filternames"]
                pupilnames = rd["pupilnames"]

            tmA.append(rd["totmag_A"])
            tmB.append(rd["totmag_B"])
            tcA.append(rd["totcts_A"])
            tcB.append(rd["totcts_B"])

        tmA = np.array(tmA)
        tmB = np.array(tmB)
        tcA = np.array(tcA)
        tcB = np.array(tcB)

        if (self.modules == "ALL") | (self.modules == "BOTH"):
            tms = [tmA, tmB]
            tcs = [tcA, tcB]
        else:
            if self.modules[0] == "A":
                tms = [tmA]
                tcs = [tcA]
            elif self.modules[0] == "B":
                tms = [tmB]
                tcs = [tcB]

        if draw_reports == True:
            nexpspec = np.max(np.array([tcA.shape[1], tcB.shape[1]]))
            nmodules = len(self.SRlist)

            f, ax = plt.subplots(
                3,
                nmodules,
                figsize=(5 * nmodules, 6),
                sharex=True,
                sharey="row",
                squeeze=False,
            )

            j_m = np.empty([attitudes.size, nmodules])
            h_m = np.empty([attitudes.size, nmodules])
            k_m = np.empty([attitudes.size, nmodules])
            DNs = np.empty([attitudes.size, nexpspec, nmodules])
            for i, att in enumerate(attitudes):
                for j in range(nmodules):
                    j_m[i, j] = tms[j][i]["j_m"]
                    h_m[i, j] = tms[j][i]["h_m"]
                    k_m[i, j] = tms[j][i]["k_m"]
                    DNs[i, :, j] = tcs[j][i]

            for j in range(nmodules):
                ax[0, j].plot(attitudes, j_m[:, j])
                ax[1, j].plot(attitudes, h_m[:, j])
                ax[2, j].plot(attitudes, k_m[:, j])

            ax[0, 0].set_ylabel("j_m")
            ax[1, 0].set_ylabel("h_m")
            ax[2, 0].set_ylabel("k_m")

            for k in range(3):
                ax[k, 0].set_ylim(24, 12)

            ax[0, 0].set_xlim(0, 360)
            for j in range(nmodules):
                ax[2, j].set_xlabel("V3_PA")
                ax[0, j].set_title("Module {}".format(self.SRnames[j]))

            f.tight_layout()

            if save_figures_dir is not None:
                f.savefig(
                    save_figures_dir
                    + "PID{}_obsid{}_mag_sweep.pdf".format(self.program, self.obsid)
                )

            nexpspec = np.max(np.array([tcA.shape[1], tcB.shape[1]]))
            prop_cycle = plt.rcParams["axes.prop_cycle"]
            colors = prop_cycle.by_key()["color"]

            if background_params is not None:
                fi = filter_info()
                zpc = zero_point_calc()
                bg = []
                ra = self.target_ra
                dec = self.target_dec
                for j in range(nexpspec):
                    wave = fi.get_info(pupilnames[j], filternames[j])
                    PHOTMJSR = zpc.get_avg_quantity(
                        pupilnames[j], filternames[j], quantity="PHOTMJSR"
                    )
                    bg.append(jbt.background(ra, dec, wave))

                res = []

                for bp in background_params:
                    below_threshold = np.ones_like(DNs, dtype=np.bool_)

                    for k in range(nexpspec):
                        bck_lambda = (
                            bp["threshold"]
                            * bp["func"](bg[k].bathtub["total_thiswave"])
                            / PHOTMJSR
                            * 1000.0
                        )
                        for j in range(nmodules):
                            below_threshold[:, k, j] = (
                                DNs[:, k, j] * 1000.0 / tot_exp_dur[k] < bck_lambda
                            )

                    below_threshold = np.all(below_threshold, axis=(1, 2))
                    V3PA_validranges_obs_starts = []
                    V3PA_validranges_obs_ends = []

                    change = np.where(below_threshold[:-1] != below_threshold[1:])[0]
                    if change.size > 0:
                        if below_threshold[change[0]]:
                            change = np.roll(change, 1)

                        V3PA_validranges_obs_starts = attitudes[change[::2]]
                        V3PA_validranges_obs_ends = attitudes[change[1::2]]
                    else:
                        V3PA_validranges_obs_starts = None
                        V3PA_validranges_obs_ends = None
                    if verbose == True:
                        print(
                            "{:3.1f} x {}(bkg)".format(
                                bp["threshold"], bp["func"].__name__
                            )
                        )
                        for s, e in zip(
                            V3PA_validranges_obs_starts, V3PA_validranges_obs_ends
                        ):
                            print(s, e)

                    res.append(
                        {
                            "s": V3PA_validranges_obs_starts,
                            "e": V3PA_validranges_obs_ends,
                            "bt": below_threshold,
                        }
                    )

            if save_report_dir is not None:
                for r, bp in zip(res, background_params):
                    filenm = "PID{}_obsid{}_report_thr{}_{}.txt".format(
                        self.program, self.obsid, bp["threshold"], bp["func"].__name__
                    )
                    with open(save_report_dir + filenm, "w") as the_file:
                        the_file.write(
                            "*** Valid ranges for PID: {}, obsid:{} ****\n".format(
                                self.program, self.obsid
                            )
                        )
                        if r["s"] is not None:
                            for s, e in zip(r["s"], r["e"]):
                                the_file.write(
                                    "PA Start -- PA End: {} -- {}\n".format(s, e)
                                )
                        else:
                            the_file.write(
                                "PA Start -- PA End: {} -- {}\n".format(0.0, 360.0)
                            )

            f, ax = plt.subplots(
                nexpspec,
                nmodules,
                figsize=(5 * nmodules, nexpspec * 2),
                sharex=True,
                sharey=True,
                squeeze=False,
            )

            for k in range(nexpspec):
                for j in range(nmodules):
                    ax[k, j].plot(attitudes, DNs[:, k, j] * 1000.0 / tot_exp_dur[k])
                    if background_params is None:
                        ax[k, j].axhline(
                            1, linestyle="dashed", label="1DN/pix/ks", c=colors[1]
                        )

                    else:
                        for l, bp in enumerate(background_params):
                            bck_lambda = (
                                bp["threshold"]
                                * bp["func"](bg[k].bathtub["total_thiswave"])
                                / PHOTMJSR
                                * 1000.0
                            )
                            ax[k, j].axhline(
                                bck_lambda,
                                linestyle="dashed",
                                label="{:3.1f} x {}(bkg) = {:5.1f} DN/pix/ks".format(
                                    bp["threshold"], bp["func"].__name__, bck_lambda
                                ),
                                c=colors[1 + l],
                            )
                            x2p = np.copy(attitudes)
                            y2p = np.copy(DNs[:, k, j])
                            x2p[res[l]["bt"]] = np.nan
                            y2p[res[l]["bt"]] = np.nan
                            ax[k, j].plot(
                                x2p, y2p * 1000.0 / tot_exp_dur[k], c=colors[1 + l]
                            )
                    ax[k, j].legend()

                ax[k, 0].set_ylabel("DN/pix/ks ({})".format(filtershort_all[k]))
                ax[k, 0].set_ylim(0.005, 500)
                ax[k, 0].set_yscale("log")

            ax[0, 0].set_xlim(0, 360)
            for j in range(nmodules):
                ax[0, j].set_title("Module {}".format(self.SRnames[j]))
                ax[-1, j].set_xlabel("V3_PA")

            f.tight_layout()
            if save_figures_dir is not None:
                f.savefig(
                    save_figures_dir
                    + "PID{}_obsid{}_DN_sweep.pdf".format(self.program, self.obsid)
                )

        return tms, tcs

    def fixed_angle(
        self,
        att,
        RP_padding=0.0,
        smooth=None,
        draw_allexp=True,
        draw_summary=True,
        nrows=2,
        ncols=3,
        savefilenames=[None, None],
        metainfo=None,
    ):
        """
        Method to check a single angle and return diagnistic plot and info.
        Useful for checking executed observations.

        Parameters
        ----------
        att: float
            position angle to check, in degrees

        RP_padding: float
            Extra padding around the susceptibility region (stars outside the nominal
            SR, but within RP_padding are flagged as "inside")

        draw_allexp: boolean
            flag to enable/disable plotting of the per-exposure results

        draw_summary: boolean
            flag to enable/disable plotting of the overall results

        nrows, ncols: integers
            number of rows and columns in the grid plot

        savefilename: [file path,file path]
            saves the figures to these location
            (first element for per-expsoure, second element for global plots).
            If None it doesn't save one file

        Returns
        ---------
        report_dict:
            dictionary containing some metinfo and a pands dataframe with the True/False
            conditions for each stars for each exposure for each susceptibility region

        """

        filtershort_all = []
        tot_exp_dur = []

        exp_spec_orders = self.nestable_obs["order_number"].unique()
        for exp_spec_order in exp_spec_orders:
            BM = self.exptable_obs["exposure_spec_order_number"] == exp_spec_order
            tot_exp_dur.append(
                self.exptable_obs.loc[BM, "photon_collecting_duration"]
                .values.astype(np.float_)
                .sum()
            )
            BM = self.nestable_obs["order_number"] == exp_spec_order
            filtershort_all.append(self.nestable_obs.loc[BM, "filter_short"].values[0])
        tot_exp_dur = np.asarray(tot_exp_dur, dtype=np.float_)

        efs_here = [
            ef
            for ef in self.efs
            if ef.nestable_row["filter_short"].values[0] == filtershort_all[0]
        ]

        # It is sufficient to check all the exposures for a single filter, because all the various exposure specifications
        # will results in an identical number of exposures, with identical pointings across the whole observation

        IN_one, V2_one, V3_one, check_one = self.check_one_angle(
            att, filtershort_all[0], RP_padding=RP_padding
        )
        colors = ["tomato", "olivedrab"]

        rpi_A = rogue_path_intensity(module="A", smooth=smooth)
        rpi_B = rogue_path_intensity(module="B", smooth=smooth)

        intensities_A = rpi_A.get_intensity(V2_one, V3_one)
        intensities_B = rpi_B.get_intensity(V2_one, V3_one)

        avg_intensity_A = np.mean(intensities_A, axis=1)
        avg_intensity_B = np.mean(intensities_B, axis=1)

        avg_V2 = np.mean(V2_one, axis=1)
        avg_V3 = np.mean(V3_one, axis=1)
        zp = 17

        emp_zp = emp_zero_point()
        zp_vega_all_A = []
        zp_vega_all_B = []
        ground_bands = []

        filternames = []
        pupilnames = []
        for filtershort in filtershort_all:
            if "+" in filtershort:
                splt = filtershort.split("+")
                pupilshort = splt[0]
                filtershort = splt[1]
            elif "_" in filtershort:
                splt = filtershort.split("_")
                pupilshort = splt[0]
                filtershort = splt[1]
            else:
                pupilshort = "CLEAR"
            zp_vega_all_A.append(emp_zp.get_emp_zp("A", pupilshort, filtershort))
            zp_vega_all_B.append(emp_zp.get_emp_zp("B", pupilshort, filtershort))
            ground_bands.append(
                emp_zp.get_ground_band(
                    pupilshort, filtershort, catalog=self.catargs["sourcecat"]
                )
            )
            filternames.append(filtershort)
            pupilnames.append(pupilshort)

        zp_vega_all_A = np.asarray(zp_vega_all_A, dtype=np.float_)
        zp_vega_all_B = np.asarray(zp_vega_all_B, dtype=np.float_)

        fict_mag_A = {}
        fict_mag_B = {}

        if self.catargs["sourcecat"] == "2MASS":
            bands = ["j_m", "h_m", "k_m"]
        else:
            bands = ["J", "H", "K"]

        for band in bands:
            fict_mag_A[band] = self.catdf[band] - 2.5 * np.log10(avg_intensity_A) + zp
            fict_mag_B[band] = self.catdf[band] - 2.5 * np.log10(avg_intensity_B) + zp
            BM_A = fict_mag_A[band] == np.inf
            fict_mag_A[band][BM_A] = 99.0
            BM_B = fict_mag_B[band] == np.inf
            fict_mag_B[band][BM_B] = 99.0

        newdf_columns = ["RAdeg", "DECdeg"] + bands
        newdf = self.catdf[newdf_columns].copy()
        newdf["avg_intensity_A"] = avg_intensity_A
        newdf["avg_intensity_B"] = avg_intensity_B
        newdf["avg_V2"] = avg_V2
        newdf["avg_V3"] = avg_V3
        for band in bands:
            newdf[band + "_fict_mag_A"] = fict_mag_A[band]
            newdf[band + "_fict_mag_B"] = fict_mag_B[band]

        totmag_A = {}
        for band in bands:
            BM = ~pd.isnull(newdf[band + "_fict_mag_A"])
            totmag_A[band] = -2.5 * np.log10(
                np.sum(np.power(10, -0.4 * newdf.loc[BM, band + "_fict_mag_A"].values))
            )
        totcts_A = []
        for i, (zp_vega, gb) in enumerate(zip(zp_vega_all_A, ground_bands)):
            totcts_A.append(10 ** ((zp_vega - totmag_A[gb]) / 2.5) * tot_exp_dur[i])

        totmag_B = {}
        for band in bands:
            BM = ~pd.isnull(newdf[band + "_fict_mag_B"])
            totmag_B[band] = -2.5 * np.log10(
                np.sum(np.power(10, -0.4 * newdf.loc[BM, band + "_fict_mag_B"].values))
            )
        totcts_B = []
        for i, (zp_vega, gb) in enumerate(zip(zp_vega_all_B, ground_bands)):
            totcts_B.append(10 ** ((zp_vega - totmag_B[gb]) / 2.5) * tot_exp_dur[i])

        report_dict = {
            "V3PA": att,
            "catalog": newdf,
            "RP_padding": RP_padding,
            "totmag_A": totmag_A,
            "totmag_B": totmag_B,
            "totcts_A": totcts_A,
            "totcts_B": totcts_B,
            "tot_exp_dur": tot_exp_dur,
            "zp_vega_all_A": zp_vega_all_A,
            "zp_vega_all_B": zp_vega_all_B,
            "filtershort_all": filtershort_all,
            "filternames": filternames,
            "pupilnames": pupilnames,
            "bands": bands,
        }

        if draw_allexp:
            f, axs = plt.subplots(nrows, ncols, figsize=(4 * ncols, 3.5 * nrows))
            for k, (ef, ax) in enumerate(zip(efs_here, axs.reshape(-1))):
                for i, (SR, SRn) in enumerate(zip(self.SRlist, self.SRnames)):
                    onepatch = patches.PathPatch(
                        SR.V2V3path, lw=2, alpha=0.5, color=colors[i], label=SRn
                    )
                    ax.add_patch(onepatch)

                if len(self.SRlist) > 1:
                    INcol = IN_one[:, 0, k] | IN_one[:, 1, k]
                else:
                    INcol = IN_one[:, 0, k]

                sz = (35.0 / (2.5 + self.catdf[self.catargs["band"]])) ** 2.0
                BM = sz > 300
                sz[BM] = 300
                BM = sz < 5
                sz[BM] = 5

                ax.scatter(V2_one[:, k], V3_one[:, k], c=INcol, s=sz)
                ax.set_xlim(3.25, -3.25)
                ax.set_ylim(7.5, 12.5)
                ax.set_xlabel("V2")
                ax.set_ylabel("V3")
                ax.legend()
                ax.set_title("Expnum: {}".format(k + 1))

            f.suptitle("Obsid: {}".format(self.obsid))
            f.tight_layout()

            if savefilenames[0] is not None:
                f.savefig(savefilenames[0])

        ###ADD PID and exptime to plot

        if draw_summary:
            f, ax = plt.subplots(1, 1, figsize=(8.5, 4))

            count_message = "Mod     Filt           t$_{{ph.c.}}$   DN/px/ks \n"
            count_message_A = None
            count_message_B = None

            for i, (SR, SRn) in enumerate(zip(self.SRlist, self.SRnames)):
                if SRn == "A":
                    label = "Mod A:"
                    for band in bands:
                        label += "\n{}$_{{tot}}$:{:6.2f}".format(band, totmag_A[band])
                    count_message_A = ""
                    for flt, t, tc in zip(filtershort_all, tot_exp_dur, totcts_A):
                        count_message_A += "  {:2}   {:11} {:5.1f}  {:6.3f}\n".format(
                            "A", flt, t, 1000.0 * tc / t
                        )

                else:
                    label = "Mod B:"
                    for band in bands:
                        label += "\n{}$_{{tot}}$:{:6.2f}".format(band, totmag_B[band])

                    count_message_B = ""
                    for flt, t, tc in zip(filtershort_all, tot_exp_dur, totcts_B):
                        count_message_B += "  {:2}   {:11} {:5.1f}  {:6.3f}\n".format(
                            "B", flt, t, 1000.0 * tc / t
                        )

                onepatch = patches.PathPatch(
                    SR.V2V3path, lw=2, alpha=0.5, color=colors[i], label=label
                )
                ax.add_patch(onepatch)

            if count_message_A is not None:
                count_message += count_message_A
            if count_message_B is not None:
                count_message += count_message_B

            if len(self.SRlist) > 1:
                IN_0 = IN_one[:, 0, :]
                IN_1 = IN_one[:, 1, :]

                IN_both = np.all(IN_0, axis=1) & np.all(IN_1, axis=1)
                IN_0_only = np.all(IN_0, axis=1) & ~(np.all(IN_1, axis=1))
                IN_1_only = np.all(IN_1, axis=1) & ~(np.all(IN_0, axis=1))

                INcol = np.all(IN_0, axis=1) | np.all(IN_1, axis=1)

                fict_mag_draw = np.empty_like(
                    newdf[self.catargs["band"] + "_fict_mag_A"]
                )
                fict_mag_draw[IN_0_only] = newdf.loc[
                    IN_0_only, self.catargs["band"] + "_fict_mag_A"
                ]
                fict_mag_draw[IN_1_only] = newdf.loc[
                    IN_1_only, self.catargs["band"] + "_fict_mag_B"
                ]
                fict_mag_draw[IN_both] = 0.5 * (
                    newdf.loc[IN_both, self.catargs["band"] + "_fict_mag_A"]
                    + newdf.loc[IN_both, self.catargs["band"] + "_fict_mag_B"]
                )

            else:
                IN_0 = IN_one[:, 0, :]
                INcol = np.all(IN_0, axis=1)
                fict_mag_draw = newdf[self.catargs["band"] + "_fict_mag_B"]

            sz = (35.0 / (-15 + fict_mag_draw)) ** 2.0
            BM = sz > 300
            sz[BM] = 300
            BM = sz < 5
            sz[BM] = 5

            ax.scatter(avg_V2, avg_V3, c=INcol, s=sz)
            ax.set_xlim(3.25, -3.25)
            ax.set_ylim(6.75, 12.75)
            ax.set_xlabel("V2")
            ax.set_ylabel("V3")
            ax.legend(
                loc=2,
                facecolor="lightgray",
                edgecolor="darkgray",
                framealpha=0.5,
                fontsize=10,
                ncol=2,
            )
            #           ax.set_title('Averaged over the {}-th exp. spec.'.format(self.nestable_obs['order_number'].values[0]),fontsize=12)

            stbox = dict(boxstyle="round", fc="lightgray", ec="darkgray", alpha=0.5)
            #           ax.text(3.1, 11.33,'{}+{}\n ZP$_{{Vega}}$={:5.2f}\n t$_{{ph.c.}} =$ {:.0f}s'.format(pupilshort,filtershort,zp_vega,tot_exp_dur), ha='left', fontsize=10,bbox=stbox)

            lcm = len(count_message.splitlines())
            fts = 10
            cmxloc, cmyloc = -3.5, 12.725

            if lcm > 10:
                fts = 9
                if lcm > 16:
                    fts = 8
                    if lcm > 20:
                        splt = count_message.splitlines()
                        hdr = splt[0]
                        hl = (lcm - 1) // 2
                        count_message = hdr[:-1] + "   " + hdr[:-1] + "\n"
                        for l in range(hl):
                            count_message += (
                                splt[1 + l][:-1]
                                + "      "
                                + splt[1 + l + hl][:-1]
                                + "\n"
                            )

            ax.text(
                cmxloc,
                cmyloc,
                count_message,
                ha="left",
                va="top",
                fontsize=fts,
                bbox=stbox,
            )

            if metainfo is not None:
                annotation = "Detectors: {}, Claws: {}".format(
                    metainfo["detector"].strip(), metainfo["claws"].strip()
                )
                for s, string in enumerate(metainfo["notes"].split(";")):
                    annotation += "\n"
                    if s == 0:
                        annotation += "Notes:"
                    annotation += string.strip()
                ax.text(3.1, 7.0, annotation, ha="left", fontsize=10, bbox=stbox)

            f.suptitle(
                "PID: {}, Obsid: {}, PA:{:6.2f}".format(self.program, self.obsid, att)
            )
            f.tight_layout()

            if savefilenames[1] is not None:
                f.savefig(savefilenames[1])

        return report_dict

    def plot_observations_checks(
        self, nrows=2, ncols=3, verbose=True, filtershort=None
    ):
        """
        Method to plot some summary results after running self.check_observations.
        It plots the claws-unaffected angles for each exposure and a summary of
        claws-unaffected angles over the whole observation

        Parameters
        ----------
        nrows, ncols: integers
            number of rows and columns in the grid plot
        """

        if filtershort is None:
            filtershort = self.nestable_obs["filter_short"].values[0]

        efs_here = [
            ef
            for ef in self.efs
            if ef.nestable_row["filter_short"].values[0] == filtershort
        ]

        #### The exposure-level plots
        f1, axs = plt.subplots(nrows, ncols, figsize=(4 * ncols, 3.5 * nrows))
        for k, (ef, ax) in enumerate(zip(efs_here, axs.reshape(-1))):
            ax.scatter(self.catdf["RAdeg"], self.catdf["DECdeg"], c="deeppink")
            ax.scatter(self.target_ra, self.target_dec, marker="X", c="red")
            ax.scatter(ef.raRef, ef.decRef, marker="X", c="orange")
            ax.axis("equal")
            ax.set_xlabel("RA")
            ax.set_ylabel("Dec")
            ax.invert_xaxis()
            ax.set_title("Expnum: {}".format(k + 1))

            for i, att in enumerate(self.attitudes):
                if self.good_angles[i, k] == True:
                    ef.define_attitude(att)
                    for SR in self.SRlist:
                        SR_RA, SR_DEC = rotations.tel_to_sky(
                            ef.attitude,
                            3600 * SR.V2V3path.vertices.T[0],
                            3600 * SR.V2V3path.vertices.T[1],
                        )
                        SR_RAdeg, SR_DECdeg = (
                            SR_RA.value * 180.0 / np.pi,
                            SR_DEC.value * 180.0 / np.pi,
                        )
                        RADEC_path = Path(
                            np.array([SR_RAdeg, SR_DECdeg]).T, SR.V2V3path.codes
                        )
                        RADEC_patch = patches.PathPatch(RADEC_path, lw=2, alpha=0.05)
                        ax.add_patch(RADEC_patch)

            draw_angstep = self.angstep
            for s, e in zip(
                self.V3PA_validranges_starts[k], self.V3PA_validranges_ends[k]
            ):
                wd = patches.Wedge(
                    (ef.raRef, ef.decRef),
                    5.5,
                    90 - e - 0.5 * draw_angstep,
                    90 - s + 0.5 * draw_angstep,
                    width=0.5,
                )
                wd.set(color="darkseagreen")

                ls = compute_line(
                    ef.raRef, ef.decRef, 90 - s + 0.5 * draw_angstep, 5.75
                )
                le = compute_line(
                    ef.raRef, ef.decRef, 90 - e - 0.5 * draw_angstep, 5.75
                )
                lm = compute_line(ef.raRef, ef.decRef, 90 - 0.5 * (s + e), 7.0)

                ax.add_artist(wd)
                ax.plot(ls[0], ls[1], color="darkseagreen")
                ax.plot(le[0], le[1], color="darkseagreen")
                ax.text(
                    lm[0][1],
                    lm[1][1],
                    "{}-{}".format(s, e),
                    fontsize=10,
                    horizontalalignment="center",
                    verticalalignment="center",
                )

            ax.set_title("Expnum: {}".format(k + 1))
        f1.suptitle("Obsid: {}".format(self.obsid))
        f1.tight_layout()

        #### The observation-level plots
        f2, ax2 = plt.subplots(1, 1, figsize=(6, 6))
        ax2.scatter(self.catdf["RAdeg"], self.catdf["DECdeg"], c="deeppink")
        ax2.scatter(self.target_ra, self.target_dec, marker="X", c="red")
        ax2.axis("equal")
        ax2.set_xlabel("RA")
        ax2.set_ylabel("Dec")
        ax2.invert_xaxis()

        draw_angstep = self.angstep
        if verbose == True:
            print("*** Valid ranges ****")

        for s, e in zip(
            self.V3PA_validranges_obs_starts, self.V3PA_validranges_obs_ends
        ):
            wd = patches.Wedge(
                (self.target_ra, self.target_dec),
                5.5,
                90 - e - 0.5 * draw_angstep,
                90 - s + 0.5 * draw_angstep,
                width=0.5,
            )
            wd.set(color="darkseagreen")

            ls = compute_line(
                self.target_ra, self.target_dec, 90 - s + 0.5 * draw_angstep, 5.75
            )
            le = compute_line(
                self.target_ra, self.target_dec, 90 - e - 0.5 * draw_angstep, 5.75
            )
            lm = compute_line(self.target_ra, self.target_dec, 90 - 0.5 * (s + e), 7.0)

            ax2.add_artist(wd)
            ax2.plot(ls[0], ls[1], color="darkseagreen")
            ax2.plot(le[0], le[1], color="darkseagreen")
            ax2.text(
                lm[0][1],
                lm[1][1],
                "{}-{}".format(s, e),
                fontsize=10,
                horizontalalignment="center",
                verticalalignment="center",
            )

            if verbose == True:
                print("PA Start -- PA End: {} -- {}".format(s, e))

        ax2.set_title("Summary for obsid {}".format(self.obsid))
        f2.tight_layout()

        return f1, f2


#### Write a method to predict the intensity of a claw based on Scott R. intensity map and
#### on the star's magnitude and exp time


class exposure_frame:
    """
    The main class to handle pointing info and rotation for an individual exposure
    """

    def __init__(self, exptable_row, target_ra, target_dec, siaf, nestable_row):
        """
        Parameters
        ----------
        exptable_row: single row in a Pandas dataframe
            contains the pointing info on this specific exposures

        siaf: instance of pysiaf.Siaf
            used to extract the info on the aperture used in this exposure

        nestable_row: single row in a Pandas dataframe
            contains info on the expsoure specification from which the exposure
             in question is generated
        """

        self.exptable_row = exptable_row
        self.nestable_row = nestable_row
        self.V2target = np.float_(self.exptable_row["v2"])
        self.V3target = np.float_(self.exptable_row["v3"])
        self.ratarget = np.float_(target_ra)
        self.dectarget = np.float_(target_dec)

    def define_attitude(self, v3pa, usetarget=True):
        """
        Define an attitude matrix (pysiaf.rotations)

        Parameters
        ----------
        v3pa:
            position angle of the v3 axis
        """
        self.attitude = rotations.attitude(
            self.V2target, self.V3target, self.ratarget, self.dectarget, v3pa
        )

    def V2V3_at_one_attitude(self, radeg, decdeg, verbose=False):
        """
        Compute V2,V3 locations of stars at a given attitude

        Parameters
        ----------
        radeg, decdeg: lists of floats
            stellar coordinates in decimal degrees

        Returns
        ---------
        v2,v3 positions in degrees

        """

        v2rads, v3rads = rotations.sky_to_tel(
            self.attitude, radeg, decdeg, verbose=verbose
        )
        return v2rads.value * 180.0 / np.pi, v3rads.value * 180.0 / np.pi


"""
Function to put together a "query by criteria" SIMBAD query 
and return an astropy Table with the results.
Query criteria here are a circle radius and a faint magnitude limit
based on a user-selectable bandpass
"""


def querysimbad(ra, dec, rad=1, band="K", maxmag=6.0, simbad_timeout=200):
    Simbad.TIMEOUT = simbad_timeout
    Simbad.reset_votable_fields()

    for filtername in ["J", "H", "K"]:
        for prop in [
            "",
            "_bibcode",
            "_error",
            "_name",
            "_qual",
            "_system",
            "_unit",
            "data",
        ]:
            field = "flux{}({})".format(prop, filtername)
            Simbad.add_votable_fields(field)

    if ra >= 0.0:
        ras = "+"
    else:
        ras = "-"
    if dec >= 0.0:
        decs = "+"
    else:
        decs = "-"

    crit = "region(circle, ICRS, {}{} {}{},{}d) & ({}mag < {})".format(
        ras, ra, decs, dec, rad, band, maxmag
    )
    print(crit)
    t = Simbad.query_criteria(crit)
    return t


"""
Function to obtain the (v1_ra,v1_dec,v3_pa) of a visit.
If a header is passed, it will get the info from it, 
if the visit_id is passed, it will query the PPSDB visit_execution table to obtain the same info
"""


def get_pointing_info(header=None, visit_id=None):
    if (header is not None) and (visit_id is not None):
        print("Cannot specify both a header and a visit_id")
        assert False

    if header is not None:
        return header["RA_V1"], header["DEC_V1"], header["PA_V3"]

    if visit_id is not None:
        pass
        # Need to write ppsdb query but need to set up the env correctly firs


## Add an option for the small SR


"""Convenience class that creates a matplotlib.Path with the Rogue Path
susceptibility zone vertices for a given NIRCam module"""


class sus_reg:
    def __init__(self, module="A", small=False):
        self.small = small
        self.module = module
        self.V2V3path = self.get_path()

    def get_path(self):
        if self.module == "A":
            if self.small == False:
                V2list = [
                    2.64057,
                    2.31386,
                    0.47891,
                    0.22949,
                    -0.04765,
                    -0.97993,
                    -0.54959,
                    0.39577,
                    0.39577,
                    1.08903,
                    1.56903,
                    2.62672,
                    2.64057,
                ]
                V3list = [
                    10.33689,
                    10.62035,
                    10.64102,
                    10.36454,
                    10.65485,
                    10.63687,
                    9.89380,
                    9.47981,
                    9.96365,
                    9.71216,
                    9.31586,
                    9.93600,
                    10.33689,
                ]
            else:
                V2list = [
                    2.28483,
                    0.69605,
                    0.43254,
                    0.57463,
                    0.89239,
                    1.02414,
                    1.70874,
                    2.28483,
                    2.28483,
                ]
                V3list = [
                    10.48440,
                    10.48183,
                    10.25245,
                    10.12101,
                    10.07204,
                    9.95349,
                    10.03854,
                    10.04369,
                    10.48440,
                ]

        else:
            if self.small == False:
                V2list = [
                    0.52048,
                    0.03549,
                    -0.28321,
                    -0.49107,
                    -2.80515,
                    -2.83287,
                    -1.58575,
                    -0.51878,
                    -0.51878,
                    -0.40792,
                    0.11863,
                    0.70062,
                    0.52048,
                ]
                V3list = [
                    10.32307,
                    10.32307,
                    10.01894,
                    10.33689,
                    10.33689,
                    9.67334,
                    9.07891,
                    9.63187,
                    8.99597,
                    8.96832,
                    9.21715,
                    9.70099,
                    10.32307,
                ]
            else:
                V2list = [
                    -0.96179,
                    -1.10382,
                    -2.41445,
                    -2.54651,
                    -2.54153,
                    -2.28987,
                    -1.69435,
                    -1.46262,
                    -1.11130,
                    -0.95681,
                    -0.59551,
                    -0.58306,
                    -0.96179,
                ]
                V3list = [
                    10.03871,
                    10.15554,
                    10.15554,
                    10.04368,
                    9.90945,
                    9.82741,
                    9.76030,
                    9.64347,
                    9.62855,
                    9.77273,
                    9.88459,
                    10.07848,
                    10.03871,
                ]

        V2list = [-1.0 * v for v in V2list]

        verts = []
        for xx, yy in zip(V2list, V3list):
            verts.append((xx, yy))
        codes = [Path.MOVETO]
        for _ in verts[1:-1]:
            codes.append(Path.LINETO)
        codes.append(Path.CLOSEPOLY)
        return Path(verts, codes)


class rogue_path_intensity:
    """
    Use the FITS files provided by Scott Rohrbach to get the intensity of the
    susceptibility zone at a given V2,V3
    """

    def __init__(self, module="A", smooth=None):
        self.module = module
        if self.module == "A":
            filename = "Rogue path NCA.fits"
        else:
            filename = "Rogue path NCB.fits"
        self.filename = "path/to/future/datadir" + filename
        self.fh = fits.getheader(self.filename)

        if smooth is not None:
            fd = fits.getdata(self.filename)
            self.fd = gaussian_filter(fd, sigma=smooth)
        else:
            self.fd = fits.getdata(self.filename)

        self.fd[:, :60] = 0.0
        self.fd[:, 245:] = 0.0
        self.fd[:85, :] = 0.0
        self.fd[160:, :] = 0.0

    def get_intensity(self, V2, V3):
        #        x = (V2-self.fh['AAXISMAX'])/(self.fh['AAXISMIN']-self.fh['AAXISMAX'])*self.fh['NAXIS1']
        x = (
            (V2 - self.fh["AAXISMIN"])
            / (self.fh["AAXISMAX"] - self.fh["AAXISMIN"])
            * self.fh["NAXIS1"]
        )
        y = (
            (V3 - self.fh["BAXISMIN"])
            / (self.fh["BAXISMAX"] - self.fh["BAXISMIN"])
            * self.fh["NAXIS2"]
        )

        xint = np.floor(x).astype(np.int_)
        yint = np.floor(y).astype(np.int_)

        BM1 = xint < 0
        BM2 = yint < 0
        BM3 = xint >= self.fh["NAXIS1"]
        BM4 = yint >= self.fh["NAXIS2"]
        BM = BM1 | BM2 | BM3 | BM4

        xint[BM] = 0
        yint[BM] = 0

        return self.fd[yint, xint]


def compute_line(startx, starty, angle, length):
    anglerad = np.pi / 180.0 * angle
    endx = startx + length * np.cos(anglerad)
    endy = starty + length * np.sin(anglerad)

    return np.array([startx, endx]), np.array([starty, endy])


class zero_point_calc:
    """
    Get the average quantities (eg zp_vega og PHOTMJSR) over SCAs for a PUPIL+FILTER combo
    """

    def __init__(self, filename="/path/to/future/datadir/" + "NRC_ZPs_0995pmap.txt"):
        df = pd.read_csv(
            filename,
            skiprows=4,
            sep="|",
            names=[
                y.strip()
                for y in "dum | pupil+filter |      sca | PHOTMJSR | zp_vega |  vega_Jy | zp_AB | mean_pix_sr | dum2".split(
                    "|"
                )
            ],
        )
        df.drop(columns=["dum", "dum2"], inplace=True)
        df["pupil+filter"] = df["pupil+filter"].map(str.strip)
        self.zp_table = df

    def get_avg_quantity(self, pupilshort, filtershort, quantity="PHOTMJSR"):
        BM = self.zp_table["pupil+filter"] == "{}+{}".format(pupilshort, filtershort)
        return np.mean(self.zp_table.loc[BM, quantity])


class emp_zero_point:
    """
    Get an empirical zp given the module and the PUPIL+FILTER combo
    """

    def __init__(self):
        self.zp_A = {
            "CLEAR+F070W": 7.0,
            "CLEAR+F090W": 8.0,
            "CLEAR+F182M": 9.7436,
            "CLEAR+F187N": 7.2,
            "CLEAR+F140M": 9.0,
            "F162M+F150W2": 7.5,
            "CLEAR+F212N": 8.1523,
            "CLEAR+F210M": 9.5981,
            "F164N+F150W2": 5.5,
            "CLEAR+F115W": 9.3934,
            "CLEAR+F150W": 10.2889,
            "CLEAR+F150W2": 10.88 - 0.37 - 0.1 - 0.0058,
            "CLEAR+F200W": 10.8321,
        }

        self.zp_B = {
            "CLEAR+F070W": 8,
            "CLEAR+F090W": 10.8491,
            "CLEAR+F182M": 10.9,
            "CLEAR+F187N": 8.3,
            "CLEAR+F140M": 9.5,
            "F162M+F150W2": 8.5,
            "CLEAR+F212N": 8.2,
            "CLEAR+F210M": 10.7,
            "F164N+F150W2": 11.0419,
            "CLEAR+F115W": 9.8801,
            "CLEAR+F150W": 11.8580,
            "CLEAR+F150W2": 12.5890,
            "CLEAR+F200W": 12.0125,
        }

        self.match2MASS = {
            "CLEAR+F070W": "j_m",
            "CLEAR+F090W": "j_m",
            "CLEAR+F182M": "h_m",
            "CLEAR+F187N": "h_m",
            "CLEAR+F115W": "h_m",
            "CLEAR+F140M": "h_m",
            "F162M+F150W2": "h_m",
            "CLEAR+F150W": "h_m",
            "CLEAR+F150W2": "h_m",
            "F164N+F150W2": "h_m",
            "CLEAR+F200W": "k_m",
            "CLEAR+F212N": "k_m",
            "CLEAR+F210M": "k_m",
        }

        self.matchSIMBAD = {
            "CLEAR+F090W": "I",
            "CLEAR+F090W": "J",
            "CLEAR+F182M": "H",
            "CLEAR+F187N": "H",
            "CLEAR+F140M": "H",
            "F162M+F150W2": "H",
            "CLEAR+F115W": "H",
            "CLEAR+F150W": "H",
            "CLEAR+F150W2": "H",
            "F164N+F150W2": "H",
            "CLEAR+F200W": "K",
            "CLEAR+F212N": "K",
            "CLEAR+F210M": "K",
        }

    def get_emp_zp(self, module, pupilshort, filtershort):
        if module == "A":
            return self.zp_A["{}+{}".format(pupilshort, filtershort)]
        elif module == "B":
            return self.zp_B["{}+{}".format(pupilshort, filtershort)]
        else:
            print("Wrong module value", module)
            assert False

    """
    This method returns the 2MASS/Simbad band to be associated with each SW NIRCam band
    to find the appropriate magnitude values for estimating counts
    """

    def get_ground_band(self, pupilshort, filtershort, catalog="2MASS"):
        if catalog == "2MASS":
            return self.match2MASS["{}+{}".format(pupilshort, filtershort)]
        elif catalog == "SIMBAD":
            return self.matchSIMBAD["{}+{}".format(pupilshort, filtershort)]


class filter_info:
    """
    Get info on filter wavelengths and bandpass
    """

    def __init__(self, filename="/path/to/future/data_dir/" + "Filter_info.txt"):
        self.filter_table = pd.read_csv(filename, sep="\s+")

    def get_info(self, pupilshort, filtershort, key_info="Pivot"):
        if pupilshort == "CLEAR":
            check_value = filtershort
        else:
            check_value = pupilshort
        BM = self.filter_table["Filter"] == check_value
        return self.filter_table.loc[BM, key_info].values[0]



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import stsci_rtd_theme

project = 'JWST Rogue Path Tool'
copyright = '2024, Mario Gennaro, Mees B. Fix'
author = 'Mario Gennaro, Mees B. Fix'
release = '0.0.1'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ['sphinx.ext.autodoc']

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'stsci_rtd_theme'
html_static_path = ['_static']
html_theme_path = [stsci_rtd_theme.get_html_theme_path()]



"""Utility funcitons for JWST Rogue Path Tool

Authors
-------
    - Mario Gennaro
    - Mees Fix
"""

from jwst_backgrounds import jbt
import numpy as np
import pandas as pd
import pathlib

from jwst_rogue_path_tool.constants import PROJECT_DIRNAME


def absolute_magnitude(band_magnitude):
    """Calculate absolute magnitude from band magnitude.

    Parameters
    ----------
    band_magnitude : float
        Band magnitude value from catalog

    Returns
    -------
    absolute_magnitude : float
        Calculated absolute magnitude of target in catalog.
    """
    absolute_magnitude = -2.5 * np.log10(band_magnitude)
    return absolute_magnitude


def calculate_background(ra, dec, wavelength, threshold):
    """Calculate background using JWST Backgrounds Tool
    """
    background_data = jbt.background(ra, dec, wavelength=wavelength, thresh=threshold)
    return background_data


def get_pupil_from_filter(filters):
    """Given a NRC filter, return list of available filters

    Parameters
    ----------
    filters : list
        List of NRC filter names

    Returns
    -------
    pupils : dict
        A dictionary of key (filter) and value (pupil)
    """
    pupils = {}

    for fltr in filters:
        if "+" in fltr:
            splt = fltr.split("+")
            pupil = splt[0]
            filter = splt[1]
        elif "_" in fltr:
            splt = fltr.split("_")
            pupil = splt[0]
            filter = splt[1]
        else:
            filter = fltr
            pupil = "CLEAR"

        pupils[filter] = pupil

    return pupils


def get_pivot_wavelength(pupil, filter):
    """Get pivot wavelength values from filter_data.txt

    Parameters
    ----------
    pupil : str
        NRC pupil name
    filter : str
        NRC filter name

    Returns
    -------
    pivot_wavelength : float
        Pivot wavelength of filter/pupil
    """
    filter_filename = pathlib.Path(PROJECT_DIRNAME) / "data" / "filter_data.txt"

    filter_table = pd.read_csv(filter_filename, sep="\s+")  # noqa

    if pupil == "CLEAR":
        check_value = filter
    else:
        check_value = pupil

    BM = filter_table["Filter"] == check_value
    pivot_wavelength = filter_table.loc[BM, "Pivot"].values[0]

    return pivot_wavelength


def get_photmjsr(pupil, filter):
    """Get photmjsr values from filter_data.txt

    Parameters
    ----------
    pupil : str
        NRC pupil name
    filter : str
        NRC filter name

    Returns
    -------
    photmjsr : float
        Flux conversion factor from DN/s to MJy/sr of filter/pupil
    """
    filter_filename = pathlib.Path(PROJECT_DIRNAME) / "data" / "filter_data.txt"

    filter_table = pd.read_csv(filter_filename, sep="\s+")  # noqa

    if pupil == "CLEAR":
        check_value = filter
    else:
        check_value = pupil

    BM = filter_table["Filter"] == check_value
    photmjsr = filter_table.loc[BM, "photmjsr"].values[0]

    return photmjsr


def make_output_directory(directory_name):
    """Make output directories for figures and text files.

    Parameters
    ----------
    directory_name : str
        Path to create directory, if exists, code will not fail but not overwrite.
    """
    pathlib.Path(directory_name).mkdir(parents=True, exist_ok=True)



#!/usr/bin/env python

import argparse
import csv
from astropy.table import Table


def arguments():
    """Parse and return command line arguments."""
    parser = argparse.ArgumentParser(
        description="Show in browser PPSDB table data (.sql) exported by APT."
        "\nList available tables, if 'table' is not specified.",
        epilog="example: aptx_sql.py 98765.sql exposures",
    )
    parser.add_argument("sqlfile", help="SQL file exported by APT")
    parser.add_argument(
        "tablename", nargs="?", default=None, help="Name of a table in the SQL file"
    )
    return parser.parse_args()


class Sqlfile:
    """An sql file exported by APT."""

    def __init__(self, sqlfile):
        """Read data from sql file. Get list of tables names."""
        self.__sqlfile = sqlfile
        self.__sql = self.sqlread()
        self.tablenames = self.tablenames()

    def sqlread(self):
        """Read sql file exported by APT. Strip trailing newlines."""
        sql = list()
        with open(self.__sqlfile, "r") as f:
            for line in f:
                sql.append(line.rstrip())
        return sql

    def tablenames(self):
        """Parse sql insert statements to determine table names."""
        prefix = "insert into "
        names = list()
        for line in self.__sql:
            if line[: len(prefix)] == prefix:
                names.append(line[len(prefix) : line.find("(")].strip())
        names = sorted(list(set(names)))
        names.remove("#AOK values")
        return names

    def rows_from_sql(self, tablename):
        """Return dictionary for each row in the specified table.
        Dictionary keys may differ for each sql insert statement.
        """
        prefix = "insert into " + tablename + " "
        rows = list()
        for line in self.__sql:
            if line[: len(prefix)] == prefix:
                keyval_str = line[len(prefix) :].strip()
                keystr, valstr = keyval_str.split("values")
                keys = [k.strip() for k in keystr[2:-2].split(",")]
                vals = [v.strip() for v in valstr[2:-2].split(",")]
                keyval_dict = dict(zip(keys, vals))
                rows.append(keyval_dict)
        return rows

    def keys(self, rows):
        keys = set()
        for row in rows:
            keys = keys.union(row.keys())
        keys = sorted(list(keys))
        return keys

    def cols_from_rows(self, rows, keys):
        table = Table()
        for key in keys:
            col = list()
            for row in rows:
                col.append(row.get(key, ""))
            try:
                col = [int(x) for x in col]
            except ValueError:
                try:
                    col = [float(x) for x in col]
                except ValueError:
                    col = [
                        x[1:-1] if x.startswith("'") and x.endswith("'") else x
                        for x in col
                    ]
            table[key] = col
        return table

    def table(self, tablename, browser=False):
        """Construct astropy table from sql insert statements.
        For the 'exposures' table, discard rows with apt_label == 'BASE'.
        Convert column data type to integer or float, where possible.
        Strip beginning and ending single quote from strings.
        """
        rows = self.rows_from_sql(tablename)
        keys = self.keys(rows)
        if len(keys) == 0:
            raise Exception("no '" + tablename + "' table in " + self.__sqlfile)
        table = self.cols_from_rows(rows, keys)
        if browser:
            self.browser(table)
        return table

    def browser(self, table):
        """Diplay copy of astropy table in a browser window.
        Convert underscores to spaces in column headers to allow wrapping.
        """
        out = table.copy(copy_data=False)
        for key in out.keys():
            newkey = key.replace("_", " ")
            if newkey != key:
                out.rename_column(key, newkey)
        out.show_in_browser(jsviewer=True, show_row_index=False)


def main():
    args = arguments()
    sql = Sqlfile(args.sqlfile)
    if args.tablename:
        table = sql.table(args.tablename, browser=True)
    else:
        print("specify a table name as the second argument:")
        for name in sql.tablenames:
            print(name)


if __name__ == "__main__":
    main()



"""Test `jwst_rogue_path_tool.utils` module.

Authors
-------
    - Mees Fix
"""

import pytest

from jwst_rogue_path_tool.utils import absolute_magnitude, get_pivot_wavelength


@pytest.mark.parametrize(
    "magnitude, expected",
    [
        (7.9, -2.2440677282261037),
        (15.6, -2.9828114958861542),
    ],
)
def test_absolute_magnitude(magnitude, expected):
    pivot_wavelength = absolute_magnitude(magnitude)
    pivot_wavelength == expected


@pytest.mark.parametrize(
    "pupil, filter, expected",
    [
        ("CLEAR", "F070W", 0.704),
        ("F070W", "F140M", 0.704),
        ("CLEAR", "F162M", 1.626),
        ("F200W", "F210M", 1.990),
    ],
)
def test_get_pivot_wavelength(pupil, filter, expected):
    pivot_wavelength = get_pivot_wavelength(pupil, filter)
    pivot_wavelength == expected



"""Unit tests for apt_json_parser"""

import pytest

from pathlib import Path

from jwst_rogue_path_tool.program_data_parser import aptJsonFile
from jwst_rogue_path_tool.constants import PROJECT_DIRNAME

JSON_PATH = Path(PROJECT_DIRNAME) / "data" / "APT_test_4RPtool.records.json"


def test_json_exists(json_path=JSON_PATH):
    """Assert json test file exists"""
    assert json_path.exists()


def test_load_json(json_path=JSON_PATH):
    program = aptJsonFile(json_path)
    assert program is not None


@pytest.mark.parametrize(
    "tablename",
    [
        "exposures",
        "visit",
        "observation",
        "program",
        "fixed_target",
    ],
)
def test_tables_exist(tablename, json_path=JSON_PATH):
    program = aptJsonFile(json_path)
    assert tablename in program.tablenames



"""
This module contains objects that perform the data organizing and analysis
routines for detecting "Claw" anomalies with NIRCam.

The AptProgram class accepts an APT JSON file that organizes data with observations,
visits, exposures and parses them into python objects (mainly pandas dataframes).

These data are then used to calculate whether exposures and observations are susceptible
to claw anomalies. We display these data in figures at the exposure and observaiton level.

Authors
-------
    - Mario Gennaro
    - Mees Fix
"""

import collections
from copy import deepcopy
from itertools import chain, groupby
import operator
import os

from astropy.io import fits
from matplotlib.path import Path
import numpy as np
import pandas as pd
import pathlib
from pysiaf.utils import rotations
from scipy.ndimage import gaussian_filter
from tqdm import tqdm

from jwst_rogue_path_tool.program_data_parser import aptJsonFile
from jwst_rogue_path_tool.constants import (
    PROJECT_DIRNAME,
    SUSCEPTIBILITY_REGION_FULL,
    SUSCEPTIBILITY_REGION_SMALL,
)
from jwst_rogue_path_tool.fixed_angle import fixedAngle
from jwst_rogue_path_tool.plotting import (
    create_exposure_plots,
    create_observation_plot,
    create_v3pa_vs_flux_plot,
)
from jwst_rogue_path_tool.utils import (
    calculate_background,
    get_pupil_from_filter,
    get_pivot_wavelength,
    get_photmjsr,
    make_output_directory,
)


class aptProgram:
    """Class that handles the APT-program-level information.
    aptProgram takes the JSON file input and uses the "observations" and "exposureFrame"
    objects to organize data into python objects that can be used for various analyses.
    """

    def __init__(
        self,
        jsonfile,
        angular_step=1.0,
        usr_defined_obs=None,
        usr_defined_angles=None,
        bkg_params=[
            {"threshold": 0.1, "function": np.min},
            {"threshold": 0.2, "function": np.mean},
        ],
    ):
        """
        Parameters
        ----------
        jsonfile : str
            Path to an APT-exported JSON file

        angular_step : float
            Attitude angle step size used to check if surrounding targets land
            in susceptibility region (default: 1.0)

        usr_defined_obs : list
            List of specific oberservations to load from program

        usr_defined_angles : list
            List of angles to valid, only the angle values in this list will be
            evaluated

        bkg_params : dictionary
            Dictionary of background threshold values and function to apply to background
            calculations. "{threshold": 0.1, "function: np.min}" will calculate the value of 10%
            of the minimum background based on JWST pointing. Docs for jwst_backgrounds
            https://github.com/spacetelescope/jwst_backgrounds
        """

        self.json = aptJsonFile(jsonfile)
        if "fixed_target" not in self.json.tablenames:
            raise Exception("JWST Rogue Path Tool only supports fixed targets")

        self.angular_step = angular_step
        self.usr_defined_obs = usr_defined_obs

        # Users may only want to check a list of specific angles of attitude,
        # if they provide a list of angles, those are the only angles that are checked.
        if angular_step and usr_defined_angles:
            raise Exception(
                "You defined an angular step that will check all valid angles (0 --> 359.0) \
                in steps of {angular_step} and usr_defined_angles of {usr_defined_angles} which \
                only checks for targets at these angles of attitude. Please select which you \
                prefer, but both options can not be true."
            )
        elif self.angular_step:
            self.angles_list = np.arange(0, 360.0, self.angular_step)
        elif usr_defined_angles:
            self.angles_list = usr_defined_angles
        else:
            raise Exception(
                "angular_step and usr_defined_angs are both None. You must pass angular_step \
                to check for targets in attitude angles (0.0 --> 359.0) in steps of `angular_step` \
                or a list of individual angles to check `usr_defined_angles` i.e. [20.0, 39.0, 256.0]"
            )
        self.bkg_parameters = bkg_params

    def __build_observations(self):
        """Convenience method to build Observation objects"""
        self.observations = observations(self.json, self.usr_defined_obs)

    def __build_exposure_frames(self):
        """Convenience method to build ExposureFrame objects"""
        for observation_id in self.observations.observation_number_list:
            if observation_id in self.observations.unusable_observations:
                continue
            else:
                observation = self.observations.data[observation_id]
                exposure_frames = exposureFrames(
                    observation, self.angles_list, self.angular_step
                )
                self.observations.data[observation_id]["exposure_frames"] = (
                    exposure_frames
                )

    def __calculate_averages(self):
        """Convenience method to average out the intensities and positions
        for an observation.
        """
        print("... Averaging intensities per angle ...")
        for observation_id in self.observations.observation_number_list:
            if observation_id in self.observations.unusable_observations:
                continue
            else:
                averages = collections.defaultdict(dict)
                observation = self.observations.data[observation_id]
                swept_angles = observation["exposure_frames"].swept_angles
                exposure_frame_ids = observation["exposure_frames"].data.keys()
                modules = observation["exposure_frames"].susceptibility_region.keys()
                for angle in self.angles_list:
                    for module in modules:
                        for quantity in ["intensity", "v2", "v3"]:
                            averages[angle][f"avg_{quantity}_{module}"] = np.mean(
                                [
                                    swept_angles[exp_frm][angle][f"{quantity}_{module}"]
                                    for exp_frm in exposure_frame_ids
                                ],
                                axis=0,
                            )
                        observation[f"averages_{module}"] = averages

    def __get_flux_vs_angle(self):
        """Get all flux values as function of angle"""
        print("... Calculating flux values ...")
        for observation_id in self.observations.observation_number_list:
            if observation_id in self.observations.unusable_observations:
                continue
            else:
                final_fluxes = collections.defaultdict(dict)
                observation = self.observations.data[observation_id]

            total_exposure_durations = observation[
                "exposure_frames"
            ].total_exposure_duration_table

            observation["filters"] = total_exposure_durations["filters"]
            observation["pupils"] = total_exposure_durations["pupils"]

            counts_per_filter = [
                fixedAngle(observation, angle).total_counts
                for angle in self.angles_list
            ]
            flux_keys = list(
                set(chain.from_iterable(sub.keys() for sub in counts_per_filter))
            )

            for key in flux_keys:
                flux_per_key = list(map(operator.itemgetter(key), counts_per_filter))
                final_fluxes["total_counts"][key] = np.array(flux_per_key)

                for filter in observation["filters"]:
                    exposure_duration = total_exposure_durations.loc[
                        (total_exposure_durations["filters"] == filter)
                    ]["photon_collecting_duration"].values[0]
                    if filter in key:
                        pix_dn_ks = (
                            final_fluxes["total_counts"][key] * 1000 / exposure_duration
                        )
                        flux_key = key.replace("total_counts", "dn_pix_ks")
                        final_fluxes["dn_pix_ks"][flux_key] = pix_dn_ks

            observation["flux"] = final_fluxes

    def __find_background_thresholds(self):
        """Find fluxes that are higher than threshold values."""
        print("... Calculating backgrounds and thresholds ...")
        for observation_id in self.observations.observation_number_list:
            if observation_id in self.observations.unusable_observations:
                continue
            else:
                observation = self.observations.data[observation_id]
                susceptibility_regions = observation[
                    "exposure_frames"
                ].susceptibility_region
                modules = susceptibility_regions.keys()
                fluxes = observation["flux"]["dn_pix_ks"]  # Get flux dictionary
                total_exposure_duration_table = observation[
                    "exposure_frames"
                ].total_exposure_duration_table
                flux_boolean = collections.defaultdict(dict)
                for parameters in self.bkg_parameters:
                    above_threshold = []
                    threshold = parameters["threshold"]
                    statistic_function = parameters["function"]
                    # Calculate the background for each filter/pupil combination
                    for module in modules:
                        for index, row in total_exposure_duration_table.iterrows():
                            pupil = row["pupils"]
                            filter = row["filters"]
                            pivot_wavelength = get_pivot_wavelength(pupil, filter)
                            photmjsr = get_photmjsr(pupil, filter)
                            background = calculate_background(
                                self.ra, self.dec, pivot_wavelength, threshold
                            )

                            wavelengths = background.bathtub["total_thiswave"]
                            lam_thresh = (
                                threshold
                                * statistic_function(wavelengths)
                                / photmjsr
                                * 1000.0
                            )

                            flux_key = f"dn_pix_ks_{pupil}+{filter}_{module}"

                            flux_above_limit = np.copy(fluxes[flux_key])
                            limit_indices = flux_above_limit < lam_thresh

                            flux_boolean_key = f"{filter}_{module}"
                            statistics_key = f"{statistic_function.__name__}_{lam_thresh}_{threshold}"
                            flux_boolean[flux_boolean_key][statistics_key] = (
                                limit_indices
                            )

                            above_threshold.append(limit_indices)

                        # Convert list to an array and then see where all indices are true
                        # for each filter/pupil combination.
                        all_boolean = np.array(above_threshold).all(0)
                        module_boolean_key = (
                            f"flux_boolean_{statistic_function.__name__}_{module}"
                        )
                        flux_boolean[module_boolean_key] = all_boolean

            observation["flux_boolean"] = flux_boolean

    def get_target_information(self):
        """Obtain RA and Dec of target from APT JSON file"""
        target_info = self.json.build_dataframe("fixed_target")

        self.ra = target_info["ra_computed"][0]
        self.dec = target_info["dec_computed"][0]

    def plot_exposures(self, observation, output_directory=None):
        """Create plot for individual exposures for a given observation. Plot
        will contain targets defined in a specific inner and outer radius
        defined by user. Check `jwst_rogue_path_tool.plotting.create_exposure_plots`
        for more information.

        Parameters
        ----------
        observation_id : int
            Observation id number to generate figures from.
        """
        create_exposure_plots(observation, self.ra, self.dec, output_directory)

    def plot_observation(self, observation, output_directory=None):
        """Create plot at the observation level. The "observation level" is
        defined as all of the valid angles from each exposure combined. Plot
        will contain targets defined in a specific inner and outer radius
        defined by user. Check `jwst_rogue_path_tool.plotting.create_observation_plot`
        for more information.

        Parameters
        ----------
        observation_id : int
            Observation id number to generate figures from.
        """
        create_observation_plot(observation, self.ra, self.dec, output_directory)

    def plot_v3pa_vs_flux(self, observation, output_directory=None):
        """Create plot of position angle vs flux. This plot will only work if users
        use the angular_step keyword calculate the flux for all 360 degrees of attitude.
        """

        if self.angular_step:
            create_v3pa_vs_flux_plot(observation, output_directory)
        else:
            raise Exception(
                "PLOTTING V3PA VS FLUX ONLY WORKS FOR OBSERVATIONS WITH ALL 360 DEGREES \
                OF ATTITUDE. SET ANGULAR STEP AND RE-RUN JWST ROGUE PATH TOOL."
            )

    def run(self):
        """Convenience method to build AptProgram"""
        self.get_target_information()
        self.__build_observations()
        self.__build_exposure_frames()
        self.__calculate_averages()
        self.__get_flux_vs_angle()
        self.__find_background_thresholds()

    from itertools import groupby

    def make_background_report(self, observation, output_directory):
        """Write reports for background restricted position angles (PA's).
        This will create a report for the module(s) in the observation provided, this
        means that the resulting PA's will be PA's that are flagged across
        all filters in the observaion for that module.

        observation : dict
            Observation out of observations.data attribute

        output_directory : str
            Output directory path
        """
        obs_id = observation["visit"]["observation"].values[0]
        program = observation["nircam_templates"]["program"].values[0]
        modules = observation["exposure_frames"].susceptibility_region.keys()
        functions = [params["function"].__name__ for params in self.bkg_parameters]
        thresholds = [params["threshold"] for params in self.bkg_parameters]

        path = pathlib.Path(output_directory)
        path = path / str(program) / "valid_angle_reports"
        make_output_directory(path)

        if path.is_dir():
            if len(modules) > 1:
                data = {}
                filename = (
                    f"program_{program}_observation_{obs_id}_background_module_A_B.txt"
                )
                full_file_path = path / filename
                f = open(full_file_path, "w")
                for threshold, function in zip(thresholds, functions):
                    data["A"] = observation["flux_boolean"][
                        f"flux_boolean_{function}_{'A'}"
                    ].nonzero()[0]

                    data["B"] = observation["flux_boolean"][
                        f"flux_boolean_{function}_{'B'}"
                    ].nonzero()[0]

                    valid_pa_angles = np.intersect1d(data["A"], data["B"])

                    f.write("**** Ranges Not Impacted by Background Thresholds ****\n")
                    f.write("**** Modules A + B ****\n")
                    f.write(f"**** Ranges Under {threshold} of {function}  ****\n")
                    for _, g in groupby(
                        enumerate(valid_pa_angles), lambda k: k[0] - k[1]
                    ):
                        start = next(g)[1]
                        end = list(v for _, v in g) or [start]
                        f.write(f"PA Start -- PA End: {start} -- {end[-1]}\n")
            else:
                single_module = list(modules)[0]
                filename = f"program_{program}_observation_{obs_id}_background_module_{single_module}.txt"
                full_file_path = path / filename
                f = open(full_file_path, "w")
                for threshold, function in zip(thresholds, functions):
                    valid_pa_angles = observation["flux_boolean"][
                        f"flux_boolean_{function}_{single_module}"
                    ].nonzero()[0]

                    f.write("**** Ranges Not Impacted by Background Thresholds ****\n")
                    f.write(f"**** Module {single_module} ****\n")
                    f.write(f"**** Ranges Under {threshold} of {function}  ****\n")
                    for _, g in groupby(
                        enumerate(valid_pa_angles), lambda k: k[0] - k[1]
                    ):
                        start = next(g)[1]
                        end = list(v for _, v in g) or [start]

                        f.write(f"PA Start -- PA End: {start} -- {end[-1]}\n")
            print(f"WROTE FILE TO: {full_file_path}")
            f.close()
        else:
            raise Exception(f"CAN NOT WRITE TO {path}, NOT A VALID DIRECTORY")

    def make_report(self, observation, output_directory):
        """Display of write "observation level" report given an observation in a program.

        Parameters
        ----------
        observation : dict
            Observation out of observations.data attribute

        output_directory : str
            Output directory path
        """

        obs_id = observation["visit"]["observation"].values[0]
        program = observation["nircam_templates"]["program"].values[0]

        all_starting_angles = observation["valid_starts_angles"]
        all_ending_angles = observation["valid_ends_angles"]

        path = pathlib.Path(output_directory)
        path = path / str(program) / "valid_angle_reports"
        make_output_directory(path)

        if path.is_dir():
            filename = f"program_{program}_observation_{obs_id}_valid_angle_report.txt"

            full_file_path = path / filename

            f = open(full_file_path, "w")
            f.write(f"**** Valid Ranges for Observation {obs_id} ****\n")
            for min_angle, max_angle in zip(all_starting_angles, all_ending_angles):
                f.write(f"PA Start -- PA End: {min_angle} -- {max_angle}\n")
            f.close()
            print(f"WROTE REPORT TO {full_file_path}")

        else:
            raise Exception(f"CAN NOT WRITE TO {path}, NOT A VALID DIRECTORY")


class observations:
    """Class the organizes metadata from APT JSON file into python object.
    This object is organized by observation number and contains metadata
    associated with it.
    """

    def __init__(self, apt_json, usr_defined_obs=None):
        """
        Parameters
        ----------
        apt_json : jwst_rogue_path_tool.program_data_parser.aptJsonFile
            Parsed JSON data into python objects (pandas dataframes)

        usr_defined_obs : list
            List of specific oberservations to load from program
        """
        self.json = apt_json
        self.program_data_by_observation(usr_defined_obs)
        self.observation_number_list = self.data.keys()
        self.drop_unsupported_observations()

    def drop_unsupported_observations(self):
        """Convenience method to drop unsupported observation types. This
        method checks all observations including parallels. All metadata
        is kept and new class attribute `self.supported_observations` which is created
        to avoid confusion when processing. `self.supported_observations` are the
        only observation from a program that are analyzed by `jwst_rogue_path_tool`.
        """
        supported_templates = [
            "NIRCam Imaging",
            "NIRCam Wide Field Slitless Spectroscopy",
        ]
        self.unusable_observations = []

        for observation_id in self.observation_number_list:
            visit_table = self.data[observation_id]["visit"]
            templates = visit_table["template"]
            exposure_table = self.data[observation_id]["exposures"]

            # If any visits have unsupported templates, this will locate them
            unsupported_templates = visit_table[~templates.isin(supported_templates)]

            # If unsupported templates is empty, NRC is primary
            if unsupported_templates.empty:
                # If template_coord_parallel_1 exists in visit table, check if secondary
                # contains non NRC exposures, remove them
                if "template_coord_parallel_1" in visit_table:
                    aperture_names = exposure_table["AperName"]
                    nrc_visits = aperture_names.str.contains("NRC")
                    self.data[observation_id]["exposures"] = exposure_table[nrc_visits]
            elif "template_coord_parallel_1" in visit_table:
                # If NRC is not the primary, check to see if NRC is the secondary.
                parallel_templates = visit_table["template_coord_parallel_1"]
                unsupported_parallels = visit_table[
                    ~parallel_templates.isin(supported_templates)
                ]
                if unsupported_parallels.empty:
                    # If NRC is secondary, make sure to remove any exposures
                    # associated with the primary instrument
                    aperture_names = exposure_table["AperName"]
                    nrc_visits = aperture_names.str.contains("NRC")
                    self.data[observation_id]["exposures"] = exposure_table[nrc_visits]
                else:
                    self.unusable_observations.append(observation_id)
            else:
                self.unusable_observations.append(observation_id)

        # Create seperate data object with unusable observations removed.
        self.supported_observations = deepcopy(self.data)
        for observation_id in self.unusable_observations:
            self.supported_observations.pop(observation_id)

    def program_data_by_observation(self, specific_observations=None):
        """Class method to organize APT data by obsevation id

        Parameters
        ----------
        specific_observations : list
            List of observations defined by user.
        """
        program_data_by_observation_id = collections.defaultdict(dict)
        target_information = self.json.build_dataframe("fixed_target")
        for table in [
            "visit",
            "exposures",
            "nircam_exposure_specification",
            "nircam_templates",
        ]:
            df = self.json.build_dataframe(table)

            unique_obs = df["observation"].unique()

            if table == "exposures":
                df = df.loc[df["apt_label"] != "BASE"]

            if specific_observations:
                for obs in specific_observations:
                    if obs not in unique_obs:
                        raise Exception(
                            (
                                "User defined observation: '{}' not available! "
                                "Available observations are: {}".format(obs, unique_obs)
                            )
                        )
                    else:
                        continue

            if specific_observations:
                observations_list = specific_observations
            else:
                observations_list = unique_obs

            for observation_id in observations_list:
                df_by_program_id = df.loc[df["observation"] == observation_id]
                program_data_by_observation_id[observation_id][table] = df_by_program_id

                program_data_by_observation_id[observation_id]["ra"] = (
                    target_information["ra_computed"].values[0]
                )

                program_data_by_observation_id[observation_id]["dec"] = (
                    target_information["dec_computed"].values[0]
                )

        self.data = program_data_by_observation_id


class exposureFrames:
    """Class the organizes data from a single observation (made of exposures)
    into exposure frames. An exposure frame is a group of exposures associated
    with a value in the NRC order specification table. Exposures with the same
    order number are a part of the same exposure frame. This object contains
    """

    def __init__(self, observation, attitudes, angular_step=None):
        """
        Parameters
        ----------
        observation : dict
            Dictionary containing data from a single observation

        attitudes : list like
            Angles of attitude to check for targets falling in the susceptibility
            region.
        """
        self.assign_catalog()
        self.observation = observation
        self.attitude_angles = attitudes
        self.angular_step = angular_step
        self.observation_number = self.observation["visit"]["observation"].values[0]
        self.exposure_table = self.observation["exposures"]
        self.template_table = self.observation["nircam_templates"]
        self.nrc_exposure_specification_table = self.observation[
            "nircam_exposure_specification"
        ]

        self.module_by_exposure = self.exposure_table.merge(
            self.template_table[["visit", "modules"]]
        ).set_index(self.exposure_table.index)

        self.exposure_frame_table = pd.merge(
            self.module_by_exposure,
            self.nrc_exposure_specification_table,
            left_on=["exposure_spec_order_number"],
            right_on=["order_number"],
            how="left",
        )

        self.get_total_exposure_duration()
        self.build_exposure_frames_data()
        self.check_in_susceptibility_region()

        # For sweeping all angles we will need a angular_step defined.
        # Else, we have a defined set of angles that are user defined and
        # do not need windows of visibility calculated.
        if self.angular_step:
            self.get_visibility_windows()

    def assign_catalog(self, catalog_name="2MASS"):
        """Obtain magnitude selected catalog as pandas dataframe.

        Parameters
        ----------
        catalog_name : str
            Survey name of catalog with star positions and magnitudes [options: 2MASS, SIMBAD]
        """
        catalog_names = {"2MASS": "two_mass_kmag_lt_5.csv", "SIMBAD": ""}

        if catalog_name not in catalog_names.keys():
            raise Exception(
                "AVAILABLE CATALOG NAMES ARE '2MASS' and 'SIMBAD' {} NOT AVAILABLE".format(
                    catalog_name
                )
            )

        self.catalog_name = catalog_name
        selected_catalog = catalog_names[self.catalog_name]
        full_catalog_path = os.path.join(PROJECT_DIRNAME, "data", selected_catalog)

        self.catalog = pd.read_csv(full_catalog_path)

    def build_exposure_frames_data(self):
        """Use exposure table to separate data into exposure frame specific
        pandas dataframes. Resetting the index to combinations of exposure and
        order number will separate the exposures into exposures associate with
        a specific dither pointing. These tables contain exposures that all
        share the same RA and Dec.
        """
        dither_pointings = self.exposure_frame_table.dither_point_index.unique()
        exposures = self.module_by_exposure.index

        self.data = {}

        for dp, exp in zip(dither_pointings, exposures):
            self.data[exp] = self.exposure_frame_table[
                self.exposure_frame_table["dither_point_index"] == dp
            ]

    def get_total_exposure_duration(self):
        total_exposure_duration_table = (
            self.exposure_frame_table.groupby("filter_short")
            .sum()["photon_collecting_duration"]
            .reset_index()
        )

        filters = total_exposure_duration_table["filter_short"]
        pupils = get_pupil_from_filter(filters)

        total_exposure_duration_table["filters"] = pupils.keys()
        total_exposure_duration_table["pupils"] = pupils.values()

        self.total_exposure_duration_table = total_exposure_duration_table

    def get_susceptibility_region(self, exposure):
        """Based on the module of an exposure frame, create a SuceptibilityRegion
        instance.

        Parameters
        ----------
        exposure : pandas.core.series.Series
            A row from an exposure frame table
        """
        sus_reg = {}
        if exposure["modules"] == "ALL" or exposure["modules"] == "BOTH":
            sus_reg["A"] = susceptibilityRegion(module="A", smooth=5)
            sus_reg["B"] = susceptibilityRegion(module="B", smooth=5)
        elif exposure["modules"] == "A":
            sus_reg["A"] = susceptibilityRegion(module=exposure["modules"], smooth=5)
        elif exposure["modules"] == "B":
            sus_reg["B"] = susceptibilityRegion(module=exposure["modules"], smooth=5)

        return sus_reg

    def get_visibility_windows(self):
        """Method to calculate when a target has entered/exited a
        susceptibility region. This is done at the exposure and
        observation levels.
        """

        # Begin calculating exposure level angles
        self.valid_starts_indices = {}
        self.valid_ends_indices = {}

        self.valid_starts_angles = {}
        self.valid_ends_angles = {}

        angles_bool_obs = []
        for exp_num in self.swept_angles:
            angles_bool = [
                self.swept_angles[exp_num][angle]["targets_in"][0]
                for angle in self.swept_angles[exp_num]
            ]

            angles_bool_obs.append(angles_bool)

            change = np.where(angles_bool != np.roll(angles_bool, 1))[0]
            # Include cases where target is in susceptibility regions
            # for every attitude angle.
            if change.size == 0:
                self.valid_starts_angles[exp_num] = None
                self.valid_ends_angles[exp_num] = None
                continue
            else:
                if angles_bool[change[0]]:
                    change = np.roll(change, 1)

            starts = change[::2]
            ends = change[1::2]

            self.valid_starts_indices[exp_num] = starts
            self.valid_ends_indices[exp_num] = ends

            self.valid_starts_angles[exp_num] = (starts - 0.5) * self.angular_step
            self.valid_ends_angles[exp_num] = (ends - 0.5) * self.angular_step

        # Begin calculating observation level angles
        angles_bool_obs = np.array(angles_bool_obs)
        angles_bool_obs = np.all(angles_bool_obs, axis=0)
        change_obs = np.where(angles_bool_obs != np.roll(angles_bool_obs, 1))[0]

        if change_obs.size == 0:
            self.observation["valid_starts_angles"] = None
            self.observation["valid_ends_angles"] = None
        else:
            if angles_bool_obs[change[0]]:
                change_obs = np.roll(change_obs, 1)

        starts = change_obs[::2]
        ends = change_obs[1::2]

        self.observation["valid_starts_angles"] = (starts - 0.5) * self.angular_step
        self.observation["valid_ends_angles"] = (ends - 0.5) * self.angular_step

    def calculate_attitude(self, v3pa):
        """Calculate attitude matrix given V3 position angle.

        Parameters
        ----------
        v3pa : float
            V3 position angle
        """
        self.attitude = rotations.attitude(
            self.exposure_data["v2"],
            self.exposure_data["v3"],
            self.exposure_data["ra_center_rotation"],
            self.exposure_data["dec_center_rotation"],
            v3pa,
        )

    def check_in_susceptibility_region(self):
        """Method to check if stars from catalog are located in susceptibility
        region per angle of attitude. Angles are 0.0 --> 360.0 degrees in steps
        of `self.angular_step`. This method creates a large dictionary that
        contains contain keys "targets_in" and "targets_loc" for each angle of
        attitude.

        For a given angle of attitude, if targets from the catalog fall in the
        susceptibility region, "targets_in" will be True and "targets_loc" are
        the indicies of these stars in the catalog.

        When an exposure frame uses both modules, "targets_in" and "targets_loc"
        are two-dimensional.

                         A      B         A      B          A      B
        "targets_in" : [True, True] or [False, True] ... [False, False]
        """
        ra, dec = self.catalog["ra"], self.catalog["dec"]
        self.swept_angles = {}

        for obs_num, obs_data in self.data.items():
            self.dataframe = obs_data
            self.exposure_data = self.dataframe.iloc[0]
            self.susceptibility_region = self.get_susceptibility_region(
                self.exposure_data
            )
            attitudes_swept = collections.defaultdict(dict)

            print(
                "Sweeping angles {} --> {} for Observation: {} and Exposure: {}".format(
                    min(self.attitude_angles),
                    max(self.attitude_angles),
                    self.observation_number,
                    obs_num,
                )
            )

            # Loop through all of the attitude angles to determine if catalog targets
            # are in the the susceptibility region.
            for angle in tqdm(self.attitude_angles):
                v2, v3 = self.V2V3_at_one_attitude(ra, dec, angle)

                # If sus_reg is dictionary, both instrument modules were used,
                # we need to check both modules for catalog targets.

                # Else only one module was used, only check there.

                # NOTE when both modules are used, `target_in` is a two dimensional
                # list. [module_a, module_b] and is a one dimensional for single modules
                attitudes_swept[angle]["targets_in"] = []
                attitudes_swept[angle]["targets_loc"] = []

                for key in self.susceptibility_region.keys():
                    in_one = self.susceptibility_region[key].V2V3path.contains_points(
                        np.array([v2, v3]).T, radius=0.0
                    )

                    if np.any(in_one):
                        attitudes_swept[angle]["targets_in"].append(True)
                        attitudes_swept[angle]["targets_loc"].append(in_one)
                    else:
                        attitudes_swept[angle]["targets_in"].append(False)
                        attitudes_swept[angle]["targets_loc"].append(in_one)

                    # Since we are already getting V2 & V3, use this opportunity to get
                    # claw intensities based on attitude angle.
                    claw_intensity = self.susceptibility_region[key].get_intensity(
                        v2, v3
                    )
                    attitudes_swept[angle][f"intensity_{key}"] = claw_intensity
                    attitudes_swept[angle][f"v2_{key}"] = v2
                    attitudes_swept[angle][f"v3_{key}"] = v3
            # Store swept angles and values.
            self.swept_angles[obs_num] = attitudes_swept

    def V2V3_at_one_attitude(self, ra_degrees, dec_degrees, v3pa, verbose=False):
        """
        Compute V2,V3 locations of stars at a given attitude

        Parameters
        ----------
        ra_degrees, dec_degrees: lists of floats
            stellar coordinates in decimal degrees

        Returns
        ---------
        v2_degrees: float
            V2 position in degrees

        v3_degrees: float
            V2 position in degrees
        """

        self.calculate_attitude(v3pa)
        v2_radians, v3_radians = rotations.sky_to_tel(
            self.attitude, ra_degrees, dec_degrees, verbose=verbose
        )

        v2_degrees = v2_radians.value * 180.0 / np.pi
        v3_degress = v3_radians.value * 180.0 / np.pi

        return v2_degrees, v3_degress


class susceptibilityRegion:
    """Class that describes the JWST NRC susceptibility regions. Creates region and
    calculates intensities based on magnitude and location of target that falls in
    susceptibility region.

    Parameters
    ----------
    module : str
        Name of JWST NRC module ("A or "B")

    small : bool
        Create smaller susceptibility region (default: False)

    smooth : bool
        Smooth data with Gaussian Filter (default: False)
    """

    def __init__(self, module, small=False, smooth=False):
        if small:
            self.module_data = SUSCEPTIBILITY_REGION_SMALL
        else:
            self.module_data = SUSCEPTIBILITY_REGION_FULL

        self.module = module
        self.smooth = smooth
        self.get_intensity_map()
        self.V2V3path = self.get_path()
        self.calculate_centroid()

    def calculate_centroid(self):
        """Calculate the centroid of a susceptibility region polygons."""
        vertices = np.array(self.verts)
        num_of_vertices = len(vertices)
        self.centroid = (
            sum(vertices[:, 0]) / num_of_vertices,
            sum(vertices[:, 1]) / num_of_vertices,
        )

    def get_intensity_map(self):
        """Open intensity map reference file"""
        if self.module == "A":
            filename = "rogue_path_nrca.fits"
        elif self.module == "B":
            filename = "rogue_path_nrcb.fits"
        else:
            ValueError(
                f"{self.module} IS NOT A VALID MODULE, VALID MODULES ARE 'A' or 'B'"
            )

        full_file_path = os.path.join(PROJECT_DIRNAME, "data", filename)
        self.fits_header = fits.getheader(full_file_path)

        if self.smooth is not None:
            fits_data = fits.getdata(full_file_path)
            self.fits_data = gaussian_filter(fits_data, sigma=self.smooth)
        else:
            self.fits_data = fits.getdata(full_file_path)

    def get_intensity(self, V2, V3):
        """Calculate the intensity of claw caused by star falling in susceptibility region."""
        self.fits_data[:, :60] = 0.0
        self.fits_data[:, 245:] = 0.0
        self.fits_data[:85, :] = 0.0
        self.fits_data[160:, :] = 0.0

        x = (
            (V2 - self.fits_header["AAXISMIN"])
            / (self.fits_header["AAXISMAX"] - self.fits_header["AAXISMIN"])
            * self.fits_header["NAXIS1"]
        )
        y = (
            (V3 - self.fits_header["BAXISMIN"])
            / (self.fits_header["BAXISMAX"] - self.fits_header["BAXISMIN"])
            * self.fits_header["NAXIS2"]
        )

        xint = np.floor(x).astype(np.int_)
        yint = np.floor(y).astype(np.int_)

        BM1 = xint < 0
        BM2 = yint < 0
        BM3 = xint >= self.fits_header["NAXIS1"]
        BM4 = yint >= self.fits_header["NAXIS2"]
        BM = BM1 | BM2 | BM3 | BM4

        xint[BM] = 0
        yint[BM] = 0

        return self.fits_data[yint, xint]

    def get_path(self):
        """Calculate rogue path for plotting."""
        V2list = self.module_data[self.module][0]
        V3list = self.module_data[self.module][1]

        V2list = [-1.0 * v for v in V2list]

        self.verts = []

        for xx, yy in zip(V2list, V3list):
            self.verts.append((xx, yy))
        self.codes = [Path.MOVETO]

        for _ in self.verts[1:-1]:
            self.codes.append(Path.LINETO)

        self.codes.append(Path.CLOSEPOLY)

        return Path(self.verts, self.codes)
